#!/usr/bin/env python
from sys import exit
from random import randrange
import math
from enum import Enum

import pyrosetta
import pyrosetta.rosetta as rosetta
import pyrosetta.rosetta.protocols.loops.loop_closure.kinematic_closure as kic
import pyrosetta.rosetta.protocols.loops.loop_mover.refine as refine

class LoopBuildMod(Enum):
    Rebuild = 1
    Perturb = 2
    Refine =3

def build_loop(p, loop_begin, loop_end, mod=LoopBuildMod.Rebuild, m=5):
    """ Adpated based on PyRosetta test/T650_LoopsKIC.py """
    #Initialization
    starting_p = pyrosetta.Pose()
    starting_p.assign( p)
    loop_cut = (loop_begin+loop_end)//2
    my_loop = rosetta.protocols.loops.Loop( loop_begin, loop_end, loop_cut )
    my_loops = rosetta.protocols.loops.Loops()
    my_loops.add_loop( my_loop )
    #fold tree for loop modeling
    rosetta.protocols.loops.set_single_loop_fold_tree( p, my_loop )

    # Randomlize inital conformation / perturb initial conformations
    if mod is LoopBuildMod.Rebuild or mod is LoopBuildMod.Perturb:
        #Initialization
        MAX_KIC_BUILD_ATTEMPTS = 10000
        scorefxn_high = rosetta.protocols.loops.get_fa_scorefxn()
        scorefxn_low  = rosetta.protocols.loops.get_cen_scorefxn()
        # use the KinematicMover explicitly in centroid stage
        kic_mover = kic.KinematicMover()
        # auxiliary movers to swith between centroid and fullatom conformation
        to_centroid = rosetta.protocols.simple_moves.SwitchResidueTypeSetMover( 'centroid' )
        to_fullatom = rosetta.protocols.simple_moves.SwitchResidueTypeSetMover( 'fa_standard' )
        recover_sidechains = rosetta.protocols.simple_moves.ReturnSidechainMover( starting_p )
        task_pack = rosetta.core.pack.task.TaskFactory.create_packer_task( starting_p )
        task_pack.restrict_to_repacking()
        task_pack.or_include_current( True )
        #pack = rosetta.protocols.simple_moves.PackRotamersMover( scorefxn_high, task_pack )
        pack = rosetta.protocols.minimization_packing.PackRotamersMover(scorefxn_high, task_pack)

        # set up centroid stage line minimizer
        tol = 0.001
        min_type = "linmin"
        mm = pyrosetta.MoveMap()
        linmin_mover = rosetta.protocols.minimization_packing.MinMover( mm, scorefxn_low, min_type, tol, True )

        to_centroid.apply( p )
        # Initial closed loop
        if mod is LoopBuildMod.Rebuild:
            print( "building random loop conformation with ideal bond lengths, bond angles," )
            print( "and omega angles using KIC", end='' )
            kic_mover.set_idealize_loop_first( True )
            kic_mover.set_pivots( loop_begin, loop_cut, loop_end )
            for i in range( MAX_KIC_BUILD_ATTEMPTS ):
              print( "\n  attempt %d..." %i, end='' )
              kic_mover.apply( p )
              if kic_mover.last_move_succeeded():
                kic_mover.set_idealize_loop_first( False )
                print( "succeeded." )
                break
            if not kic_mover.last_move_succeeded():
              print( "Could not complete initial KIC loop building in %d attempts. Exiting" %MAX_KIC_BUILD_ATTEMPTS )
              exit()
            scorefxn_low( p )
            linmin_mover.apply( p )
        # Loop perturbation
        print( "centroid stage loop KIC remodeling" )
        if mod is LoopBuildMod.Rebuild:
            outer_cycles = 10 # 10 # inner_cycles and outer_cycles should be much higher in production run!
            inner_cycles = (loop_end-loop_begin-1)*5 # 30
        else:  # LoopBuildMod.Perturb
            outer_cycles = 4
            inner_cycles = (loop_end-loop_begin-1)*2
        init_temp = 2.0
        final_temp = 1.0
        gamma = math.pow( ( final_temp/init_temp ),( 1.0/( outer_cycles*inner_cycles ) ) )
        kT = init_temp
        mc = pyrosetta.MonteCarlo( p, scorefxn_low, kT )
        for i in range( 1,outer_cycles+1 ):
            mc.recover_low( p )
            scorefxn_low( p )
            for j in range( 1, inner_cycles+1 ):
                kT = kT * gamma
                mc.set_temperature( kT )
                kic_start = randrange( loop_begin, loop_end - 1 )
                kic_end = randrange( kic_start+2, loop_end+1 )
                middle_offset = ( kic_end - kic_start ) // 2
                kic_middle = kic_start + middle_offset
                kic_mover.set_pivots( kic_start, kic_middle, kic_end )
                kic_mover.set_temperature( kT )
                kic_mover.apply( p )
                linmin_mover.apply( p )
                mc.boltzmann( p )
                rms = rosetta.protocols.loops.loop_rmsd( p, starting_p, my_loops )
                print( "centroid stage loop rmsd to starting loop: %f" % rms )
        mc.recover_low( p )
        # Return to full atom representation
        to_fullatom.apply( p )
        recover_sidechains.apply( p )
        pack.apply( p )
    #Loop refinement
    print( "high-res KIC refinement" )
    #rosetta.basic.options.set_integer_option('loops:outer_cycles', 5)  # !!! Not working (pyrosetta 150) -- but I do not want to change it anyway
    rosetta.basic.options.set_integer_option('loops:max_inner_cycles', (loop_end-loop_begin+1)*m)
    loop_refine = refine.LoopMover_Refine_KIC( my_loops )
    loop_refine.apply( p )
    # Restore the fold tree -- causes problem to fastrelax if it is not restored
    p.fold_tree(starting_p.fold_tree())


def test():
    pyrosetta.init()
    #init(extra_options='-loops:outer_cycles 5 -loops:max_inner_cycles 33')
    #init(extra_options='-loops:fast')
    p = rosetta.core.import_pose.pose_from_file( "2cpl_min.pdb.gz" )
    relax = rosetta.protocols.relax.FastRelax()
    scorefxn = pyrosetta.get_fa_scorefxn()
    relax.set_scorefxn(scorefxn)
    for i in range(2):
        working_p = pyrosetta.Pose()
        working_p.assign(p)
        build_loop(working_p, 145, 155, LoopBuildMod.Rebuild, m=5)
        working_p.dump_pdb( "kic-{:02d}rb0.pdb".format(i) )
        relax.apply(working_p)
        working_p.dump_pdb( "kic-{:02d}rb1.pdb".format(i) )
    for i in range(2):
        working_p = pyrosetta.Pose()
        working_p.assign(p)
        build_loop(working_p, 145, 155, LoopBuildMod.Perturb, m=5)
        working_p.dump_pdb( "kic-{:02d}pt0.pdb".format(i) )
        relax.apply(working_p)
        working_p.dump_pdb( "kic-{:02d}pt1.pdb".format(i) )
    for i in range(2):
        working_p = pyrosetta.Pose()
        working_p.assign(p)
        build_loop(working_p, 145, 155, LoopBuildMod.Refine, m=5)
        working_p.dump_pdb( "kic-{:02d}mt0.pdb".format(i) )
        relax.apply(working_p)
        working_p.dump_pdb( "kic-{:02d}mt1.pdb".format(i) )

if __name__ == '__main__':
    test()
