import functools
import collections
from hb_lib import seqlib, strulib

# Auxiliary functions
def combine_x(dataframes, xcol, msa, f_intra, f_inter):
    """ Combine property x from different complex and corrdinate it by site in MSA(i_msa) """
    # For each complex, coorinate asa, dASA by i_msa 
    x_list = []
    raw2aln =  dict((s.replace('-', '') , s) for s in msa)
    for df in dataframes:
        stru2msa = map_stru2msa(df, raw2aln)
        x = merge_chains(df, xcol, stru2msa, f_intra)
        x_list.append(x)
    # Combine x from multiple complexes
    result = merge_complexes(x_list, f_inter)
    return result

def map_stru2msa(dataframe, raw2aln):
    """ Map structure residue id (index of dataframe (must have column 'AA')) to i_msa """
    # Organize by chains
    chains = collections.defaultdict(list) 
    for resid in dataframe.index:
        if ';' in resid: # resid is like "(;0;A; ,276, )"
            decoded = strulib.decode_id(resid)
            cid = tuple(decoded[:3])  # SMCRA
            het_tag = decoded[3][0] 
            if het_tag != ' ': continue  # Skip het residues (MSE may also be skipped)
            chains[cid].append((resid, dataframe.loc[resid, 'AA']))
        else: # resid is like "B.110" or "B.110A"
            cid = resid[0]
            chains[cid].append((resid, dataframe.loc[resid, 'AA']))
    # Map id in structure to aligned sites in msa
    stru2msa = {}
    for cid, resid_aas in chains.items():
        # resid_aas is assumed to be ordered in the PDB structure (true in most cases, if it is not, sort it by resid)
        rawseq = ''.join(aa for resid, aa in resid_aas)
        if rawseq in raw2aln:
            aln = raw2aln[rawseq]
            raw2msa = seqlib.idmap_raw2format(aln)
            stru2msa.update((resid, raw2msa[iraw]) for iraw, (resid, aa) in enumerate(resid_aas))
    return stru2msa

def merge_complexes(x_list, f_inter):
    all_sites = functools.reduce(lambda s, a: s.union(a), (set(x.keys()) for x in x_list), set())
    result = collections.OrderedDict()
    for i_msa in sorted(all_sites):
        result[i_msa] = f_inter([x[i_msa] for x in x_list if i_msa in x])
    return result

def merge_chains(dataframe, xcol, stru2msa, f_intra):
    """ Coordinate the specified column by i_msa """
    x_imsa = collections.defaultdict(list)
    for rid in dataframe.index:
        if rid in stru2msa:
            x_imsa[stru2msa[rid]].append(dataframe.loc[rid])
    result = {}
    for i_msa, recs in x_imsa.items():
        result[i_msa] = f_intra([r[xcol] for r in recs])
    return result 

