#!/usr/bin/env python

import sys
import argparse
import collections
import functools

import numpy as np
import pandas as pd

from hb_lib import misc, seqlib, strulib
import aux


def parse():
    parser = argparse.ArgumentParser(description="Get combined weights of given list of weights", formatter_class=misc.SmartFormatter)
    required = parser.add_argument_group('required arguments')
    required.add_argument('-msa',
            help="R|The input Multiple Sequence Alignment with sequences to\n"
                 "coordinate the weights\n",
            required=True,
            type=argparse.FileType('rt'));
    required.add_argument('-wts',
            help="R|One weight file or a list of weight files\n"
                 "- If one weight file is specified, it may from a oligomer:ab \n"
                 "  complex (eg. Env:PGT145)\n"
                 "- If multiple weight files are specified, mean value of\n"
                 "  corresponding sites are used.",
            required=True, nargs='+')
    optional = parser.add_argument_group('optional arguments')
    optional.add_argument('-o', '--outfile', help='Output file (default: stdout)', default=sys.stdout, type=argparse.FileType('wt'))
    return parser.parse_args()

def main():
    para = parse()

    heads, msa = seqlib.read_fasta(para.msa)

    # Read the weight of epitope sites
    wts_collection = []
    for wt_file in para.wts:
        wts_collection.append(pd.read_table(wt_file, index_col=0, comment='#'))
    # Combine wts from different complex and corrdinate them by site in MSA (i_msa)
    weights = aux.combine_x(wts_collection, 'weight', msa, f_intra=sum, f_inter=np.mean)

    # AAs
    raw2aln =  dict((s.replace('-', '') , s) for s in msa)
    imsa2strus = [dict((imsa, stru_id) for stru_id, imsa in aux.map_stru2msa(df_wts, raw2aln).items())
                    for df_wts in wts_collection]
    all_sites = set()
    for imsa2stru in imsa2strus:
        all_sites.update(imsa2stru.keys())
    all_sites = sorted(all_sites)
    AAs = collections.defaultdict(list)
    for i_msa in all_sites: 
        for imsa2stru, df in zip(imsa2strus, wts_collection):
            try:
                stru_id = imsa2stru[i_msa]
                AAs[i_msa].append(df.loc[stru_id, 'AA'])
            except KeyError:
                AAs[i_msa].append('-')
    # Output
    fprint = functools.partial(print, file=para.outfile)
    fprint('#>{}'.format(' '.join(sys.argv)))
    fprint('i_msa AA weight'.replace(' ', '\t'))
    for i_msa in all_sites:
        if i_msa in weights:
            w = weights[i_msa]
        else:
            w = 0.0
        aa_rec = ''.join(AAs[i_msa])
        fprint(f'{i_msa}\t{aa_rec}\t{w:.3f}')

if __name__ == '__main__':
    main()
