#!/usr/bin/env python

import sys
import argparse

import Bio.PDB
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from hb_lib import misc
from hb_lib import strulib
from hb_lib import eplib
from hb_lib import nbs 

# threshold = 98%_percentile. weights_over_98%_percentile is set to threshold
th_natoms = 12.0
th_nnbs = 8.0
# scale factor to make 98%_of_glycan_terms * scale == 98%_of_aa_terms
s_natoms = 4.7
s_nnbs = 2.5
roundit = lambda x: round(x*10)/10.0
#

def parse():
    parser = argparse.ArgumentParser(description="Get weight of epitope sites according to specified Ab:Ag structures", formatter_class=misc.SmartFormatter)
    parser.add_argument('pdbfile', help='pdbfile of complex structure (.pdb, .pdb.gz, .cif, .cif.gz)')
    parser.add_argument('ag_chains', help='chain ids of antigen in the complex structure')
    parser.add_argument('ab_chains', help='chain ids of antibody in the complex structure')
    parser.add_argument('-range', help='start-end of residue id in the chimeric antigen chain (e.g. 123-194, for HIV Env in chain G of 5ESZ) (default: None)', default=None)
    parser.add_argument('--epsites', help='Print only epitope sites, instead of all sites (default: False)', default=False, action='store_true')
    parser.add_argument('-o', '--outfile', help='Output file (default: stdout)', default=sys.stdout, type=argparse.FileType('wt'))
    return parser.parse_args()

def main():
    para = parse()

    if para.range:
        pair = [int(x) for x in para.range.strip().split('-')]
        assert(len(pair) == 2)
        segment = pair
    else:
        segment = None

    parser = Bio.PDB.PDBParser()  # default: PDB format
    if '.cif' in para.pdbfile:  # Not tested yet
        parser = Bio.PDB.MMCIFParser()

    with misc.zopen(para.pdbfile) as infile:
        structure = parser.get_structure('', infile)[0]

    w_natoms = eplib.extr_epitope_natoms(structure, para.ag_chains, para.ab_chains)
    w_nnbs = get_nnbs(structure, para.ag_chains, para.ab_chains, segment, 8.5)

    w_natoms = dict((rec[0], rec[1:]) for rec in w_natoms)
    weights = {}
    for resid, nnbs_aa, nnbs_glycan, nb_ag_sites in w_nnbs:
        res = strulib.get_by_full_id(structure, resid)
        if res.resname not in misc.AA3to1:
            continue
        #
        if resid in w_natoms:
            natoms_aa, natoms_glycan = w_natoms[resid]
            #print(resid, nnbs_aa, nnbs_glycan, natoms_aa, natoms_glycan, nb_ag_sites)
        else:
            natoms_aa, natoms_glycan = 0, 0
        #
        w_natom = natoms_aa + natoms_glycan/s_natoms
        if w_natom > th_natoms:
            w_natom = th_natoms
        w_nnb = nnbs_aa + nnbs_glycan/s_nnbs
        if w_nnb > th_nnbs:
            w_nnb = th_nnbs
        #
        weights[resid] = (w_natom, w_nnb)
    #
    epsites = [resid for resid, (w_natom, w_nnb) in weights.items() if w_natom+w_nnb > 0]
    scale_w_natoms = sum(weights[rid][0] for rid in epsites) / len(epsites) * 2
    scale_w_nnbs = sum(weights[rid][1] for rid in epsites) / len(epsites) * 2
    # Output
    outs = para.outfile
    outs.write("#> {}\n".format(' '.join(sys.argv)))
    #outs.write("site\tAA\tweight\tw_natom\tw_nnb\tneighbor_sites\n")
    outs.write("site\tAA\tweight\tw_natom\tw_nnb\n")
    for resid, nnbs_aa, nnbs_glycan, nb_sites in w_nnbs:
        res = strulib.get_by_full_id(structure, resid)
        if res.resname not in misc.AA3to1:
            continue
        if para.epsites and (nnbs_aa == 0 and nnbs_glycan == 0 and resid not in w_natoms):
            continue
        #
        site = to_site(resid)
        aa = misc.AA3to1[res.resname]
        w_natom, w_nnb = weights[resid]
        weight = w_natom/scale_w_natoms + w_nnb/scale_w_nnbs
        #outs.write(f'{site}\t{aa}\t{weight:.3f}\t{w_natom:.1f}\t{w_nnb:.1f}\t{nb_sites}\n')
        outs.write(f'{site}\t{aa}\t{weight:.3f}\t{w_natom:.1f}\t{w_nnb:.1f}\n')

to_site = lambda resid: resid[2]+'.'+''.join(str(x) for x in resid[3]).replace(' ','')

def get_nnbs(model, Ag_chains, Ab_chains, segment=None, threshold=8.5, whitelist=misc.Glycans|misc.AAs):
    """ Get Ab nb sites for all ag sites """
    # Clean up the structure: remove H atoms + remove non-(AA/Glycan) residues
    strulib.clean_stru(model, whitelist)
    glycan_groups = strulib.group_glycan(model)
    glycan2ngs = dict((glycan, ngs) for ngs, glycans in glycan_groups.items() for glycan in glycans)
    glycan_chains = set(glycan[2] for ngs, glycans in glycan_groups.items() for glycan in glycans if ngs[2] in Ag_chains)
    # Get nb_dict
    all_chains = set(Ag_chains+Ab_chains) | glycan_chains
    residues = [r for r in model.get_residues() if r.parent.id in all_chains]
    #print(residues)
    nb_dict = nbs.get_delaunay_nbs(residues, threshold, scc=True)  # {res_full_id: [nb_full_id1, nb_full_id2, ...]}
    # Get result
    if segment is not None:
        start, end = segment
        within_range = lambda i: i >= start and i <= end
    note_nnbs = []
    for res in residues:
        if res.parent.id in Ab_chains: continue
        if (segment is not None) and (not within_range(res.id[1])): continue
        resid = res.get_full_id()
        # if glycosylated, count contacts with glycans
        nbs_glycan = set()
        if resid in glycan_groups:
            for glycan_id in glycan_groups[resid]:
                for nb_resid in nb_dict[glycan_id]:
                    if nb_resid[2] in Ab_chains:
                        nbs_glycan.add(nb_resid)
        nnbs_glycan = len(nbs_glycan)
        # get valid Ab+Ag neighbor sites, if segment is specified
        if segment is not None:
            all_current_nbs = []
            for rid in nb_dict[resid]:
                if rid[2] in Ab_chains:
                    all_current_nbs.append(rid)
                else:
                    if within_range(rid[3][1]):
                        all_current_nbs.append(rid)
        else:
            all_current_nbs = nb_dict[resid]
        # 
        nnbs_aa = len([rid for rid in all_current_nbs if rid[2] in Ab_chains])
        #nb_ag_sites = ','.join(to_site(rid) for rid in all_current_nbs if rid[2] in Ag_chains)
        nb_ag_sites = ','.join(to_site(rid) for rid in all_current_nbs)
        note_nnbs.append((resid, nnbs_aa, nnbs_glycan, nb_ag_sites))
    return note_nnbs

if __name__ == '__main__':
    main()
