#!/usr/bin/env python

import numpy as np
import sys
import argparse
import itertools as it

def transpose(msa):
    try:
        msa_t = []  # Transposed msa
        for j in range(len(msa[0])):
            msa_t.append(''.join(x[j] for x in msa))
    except KeyError:
        #print('Check whether the MSA is empty or not same lengh.', file=sys.stderr)
        assert(False)
    return msa_t

def eliminate_gaps(msa, th=0.2, gap_chars='-#*', triplet=False):
    """Eliminate cols have high fraction of gaps

    Return: modified_msa, the alignment with gaps removed

    Keyword arguments:
    msa -- the input alignment, [s1, s2, s3, ...]
    th -- the threshold of gap ratio, above which the column will be removed
    gap_chars -- chars in the alignmen be considered as gaps
    """
    spacer = set(gap_chars)
    n_seqs = len(msa)
    msa_t = transpose(msa)
    step = 1 if not triplet else 3
    result_t = []
    for i in range(0, len(msa_t), step):
        n_gaps = 0.0
        for j in range(n_seqs):
            is_gap = all(msa_t[i+k][j] in spacer for k in range(step))
            n_gaps += 1.0 if is_gap else 0.0
        gap_ratio = n_gaps / n_seqs
        if gap_ratio < th:
            result_t += [msa_t[i+k] for k in range(step)]
    return transpose(result_t)

def idmap_raw2aln(formated_seq, spacer='-'):
    accumulator = 0
    idmap = {}
    for i, aa in enumerate(formated_seq):
        if aa not in spacer:
            idmap[accumulator] = i
            accumulator += 1
    return idmap

def fasta_iter(filestream):
    faiter = (x[1] for x in it.groupby(filestream, lambda line: line[0] == ">"))
    for header in faiter:
        header_str = header.__next__()[1:].strip()
        seq = "".join(s.strip() for s in faiter.__next__())
        yield header_str, seq

def read_fasta(filestream):
    headers = []
    sequences = []
    for h, s in fasta_iter(filestream):
        headers.append(h)
        sequences.append(s)
    return (headers, sequences)

def write_fasta(filestream, head, seq, length=60):
    filestream.write('>{}\n'.format(head))  # filestream must be text mode ('wt')
    filestream.write('{}\n'.format(seq))

def parse():
    parser = argparse.ArgumentParser(description="Adjust Hyper-Variable Loops")
    parser.add_argument('msa', help='Muliple Sequence Alignment (fasta format, HXB2 should be included (incated by "HXB2" in sequence name))', type=argparse.FileType('rt'))
    parser.add_argument('-o', '--outfile', help='output adjusted sequences (default: stdout).', type=argparse.FileType('wt'), default=sys.stdout)
    return parser.parse_args()

def adjust(seq_row, start, end):
    """ Adjust the HV loops such that half attached to start position, half attached to end position """
    to_be_adjusted = seq_row[start:end]
    no_gaps = ''.join(to_be_adjusted).replace('-','')
    half_len = int(len(no_gaps)/2)
    adjusted = no_gaps[:half_len] + '-'*(len(to_be_adjusted)-len(no_gaps)) + no_gaps[half_len:]
    for i, aa in zip(range(start, end), adjusted):
        seq_row[i] = aa 

def adjust_half(seq_row, start, end):
    """ Adjust the HV loops - first half, left adjusted; second half, right adjusted

    Used when
    (1) there are missing sites in hypervariable loops (e.g. sequence from structure)
    (2) the inital alignment have been adjusted earlier, but gaps are introduced after the adjustment
    """
    half = int((start+end)/2)
    first = ''.join(seq_row[start:half]).replace('-','')
    second = ''.join(seq_row[half:end]).replace('-','')
    adjusted = first + '-'*(end-start-len(first)-len(second)) + second
    for i, aa in zip(range(start, end), adjusted):
        seq_row[i] = aa 

def main():
    para = parse()
    headers, seqs = read_fasta(para.msa)
    # Find hxb2 sequence
    hxb2_seq = None
    for n, s in zip(headers, seqs):
        if 'HXB2' in n.upper():
            hxb2_seq = s
    assert(hxb2_seq is not None), 'HXB2 sequence should be included in the alignment'
    num_seq = len(seqs)
    num_col = len(seqs[0])
    fmt = '{:-<%d}' %(num_col)
    seq_array = np.array([list(fmt.format(s)) for s in seqs])
    # location of HVs (HXB2 number --> MSA pos)
    hvs_hxb2 = [(131, 155), (182, 193), (391, 414), (457, 470)]  # Adjusted to include anchor sites
    idmap = idmap_raw2aln(hxb2_seq)
    hvs_msa = [(idmap[s-1], idmap[e]) for s, e in hvs_hxb2]
    for s, e in hvs_msa: print(s, e, hxb2_seq[s:e])
    # Adjust HV part
    for h, seq_row in zip(headers, seq_array):
        for s, e in hvs_msa:
            adjust(seq_row, s, e)
    #
    new_seqs = [''.join(row) for row in seq_array]
    seqs_nogap_cols = eliminate_gaps(new_seqs, th=0.9999)
    for h, s in zip(headers, seqs_nogap_cols):
        write_fasta(para.outfile, h, s)


if __name__ == "__main__":
    main()
