#!/usr/bin/env python

import sys
import argparse
import functools
import collections
import time

import numpy as np
import Bio.PDB
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from inspect import currentframe, getframeinfo
from pathlib import Path
filename = getframeinfo(currentframe()).filename
parent = Path(filename).resolve().parent.parent
sys.path.insert(0, str(parent))
from hb_lib import strulib
from hb_lib import misc


def parse():
    parser = argparse.ArgumentParser(description="Get depth and asa of each atoms")
    parser.add_argument('input_pdb', help='input_pdb (.pdb/.pdb.gz)')
    parser.add_argument('-p', '--probe_radius', help='Radius of depth probe (default: 7.0).', default=7.0, type=float)
    parser.add_argument('-e', '--extend', help='Distance extended from antibody exclusive surface asa (default: 10.0).', default=10.0, type=float)
    parser.add_argument('--include_hydrogen', help='Including hydrogen atoms, if they are included in the structure (default: False).', default=False)
    parser.add_argument('--byres', help='Return result by residue, insteat of by atoms (default: False).', action='store_true', default=False)
    parser.add_argument('-o', '--outfile', help='Output file (default: stdout).', default=sys.stdout, type=argparse.FileType('wt'))
    return parser.parse_args()

def depth_sphere(xyzs):
    # Get center
    xyz = np.array(xyzs)
    assert(xyz.shape[1] == 3)
    o = np.mean(xyz, axis=0)
    # Get distance to center 
    depth = xyz - o
    depth = np.sqrt(np.sum(depth * depth, axis=1))
    # Get outmost point and set it depth as zero
    depth = np.max(depth) - depth
    return depth

def main():
    start = time.process_time()
    para = parse()
    xyzrn = strulib.XYZRN()
    parser = Bio.PDB.PDBParser()
    with misc.zopen(para.input_pdb) as input_pdb:
        struct = parser.get_structure(para.input_pdb, input_pdb)
        if not para.include_hydrogen:  # Remove hydrogen atoms
            for residue in struct.get_residues():
                atoms_to_remove = []
                for atom in residue:
                    if atom.name.startswith('H'):
                        atoms_to_remove.append(atom)
                for atom in atoms_to_remove:
                    residue.detach_child(atom.id)
        else:
            pass
        #
        xyz, r, names = xyzrn.deduce(struct, include_het=False)
    smpl_dots = misc.smpl2048
    print('{:.3f}s: {} is read and parsed'.format(time.process_time()-start, para.input_pdb))
    asa = strulib.dots_asa(xyz, r, smpl_dots, 1.5)  # asa: ASA of atoms, corresponding to order in xyz and r
    print('{:.3f}s: dots_asa is done'.format(time.process_time()-start))

    #depth = depth_sphere(xyz)
    #print('{:.3f}s: depth_sphere is done'.format(time.process_time()-start))
    smpl_dots = misc.smpl256
    depth = strulib.dots_depth(xyz, r, smpl_dots, para.probe_radius, para.extend, [], [], remove_interior_cavities=True)

    #frame_xyz = []
    #for res in struct.get_residues():
    #    frame_xyz += [res[atom].coord for atom in 'C N O CA CB'.split() if atom in res]
    #depth = strulib.dots_depth2(frame_xyz, xyz, smpl_dots, para.probe_radius, para.extend)
    #print('{:.3f}s: dots_depth2 is done'.format(time.process_time()-start))

    output=functools.partial(print, file=para.outfile)
    if para.byres:
        depth_by_atom = collections.OrderedDict((n, d) for n, d in zip(names, depth))
        asa_by_atom = collections.OrderedDict((n, a) for n, a in zip(names, asa))
        depth_by_res = strulib.atom2res(struct, depth_by_atom, f=np.mean)
        asa_by_res = strulib.atom2res(struct, asa_by_atom, f=sum)
        #print(depth_by_res)
        output('#{}'.format(' '.join(sys.argv)))
        output('res_full_id AA depth asa rsa'.replace(' ', '\t'))
        for resid in depth_by_res:
            res = strulib.get_by_full_id(struct, strulib.decode_id(resid))
            aa = misc.AA3to1[res.get_resname()]
            d = depth_by_res[resid]
            a = asa_by_res[resid]
            output(f'{resid}\t{aa}\t{d:.2f}\t{a:.1f}\t{a/misc.TotalASA[aa]:.3f}')
    else:
        output('#{}'.format(' '.join(sys.argv)))
        output('atom_full_id depth asa'.replace(' ', '\t'))
        for n, d, a in zip(names, depth, asa):
            output(f'{n}\t{d:.2f}\t{a:.2f}')

if __name__ == "__main__":
    main()
