#!/usr/bin/env python

import sys, os
import argparse
import re

import numpy
import Bio.PDB
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from inspect import currentframe, getframeinfo
from pathlib import Path
filename = getframeinfo(currentframe()).filename
parent = Path(filename).resolve().parent.parent
sys.path.insert(0, str(parent))
from hb_lib import misc
from hb_lib import strulib
from hb_lib import seqlib


def parse():
    parser = argparse.ArgumentParser(description="Get sequence of specified chains", formatter_class=misc.SmartFormatter)
    parser.add_argument('pdbfile', help='pdbfile of input structure (.pdb, .pdb.gz)')
    parser.add_argument('-c', '--chains', help='chain ids (e.g. A, AB) (default: None)')
    parser.add_argument('-g', help="character to indicate the location of a gap (default: None)", default=None)
    parser.add_argument('-o', '--outfile', help='Output file (default: stdout)', default=sys.stdout, type=argparse.FileType('wt'))
    return parser.parse_args()

def main():
    para = parse()
    parser = Bio.PDB.PDBParser()
    min_sqdist = lambda r_a, r_b: numpy.min(strulib.get_sqdist_mat(r_a.child_list, r_b.child_list))
    nb_dist_th = 2.5 * 2.5

    with misc.zopen(para.pdbfile) as infile:
        structure = parser.get_structure('', infile)[0]
    all_chains  = ''.join(chain.id for chain in structure)
    selected_chains = para.chains if para.chains is not None else all_chains
    seq_heads, seqs = [], []
    for chain in structure:
        prefix = os.path.splitext(os.path.basename(para.pdbfile))[0]
        if chain.id not in selected_chains: continue
        seq_heads.append(f'#{prefix}_{chain.id}')
        all_residues = [r for r in chain]
        aa_list = []
        for r_a, r_b in zip(all_residues, all_residues[1:]):
            aa_list.append(seqlib.aa3to1(r_a.resname))
            if para.g is not None and min_sqdist(r_a, r_b) > nb_dist_th:
                    aa_list.append(para.g)  # Add the indicator of a missing segment
        aa_list.append(seqlib.aa3to1(all_residues[-1].resname))
        current_seq = ''.join(aa_list) #.replace('X', '@')
        #current_seq = ''.join(aa_list).replace('X', '')
        current_seq = re.sub('@+$', '', current_seq)
        current_seq = re.sub('^@+', '', current_seq)
        seqs.append(current_seq)

    # Output
    for h, s in zip(seq_heads, seqs):
        if len(s) > 0:
            seqlib.write_fasta(para.outfile, h, s)

if __name__ == '__main__':
    main()
