#include <iostream>
#include <cmath>

#include <Eigen/Core>
#include <Eigen/Geometry>
#include <Eigen/StdVector>

#include "dots_asa.h"

// TODO: parallel it

// CubicIndex
std::string CubicIndex::get_key(const Eigen::Vector3d &a) const {
  return STR(ICEIL(a[0]/grid_)) + ',' +
         STR(ICEIL(a[1]/grid_)) + ',' +
         STR(ICEIL(a[2]/grid_));
}

CubicIndex::CubicIndex(const StdVecPt &xyz, double index_grid) {
  grid_ = index_grid;
  pxyz_ = &xyz;
  FOR(i, xyz.size()) {
    std::string key = get_key(xyz[i]);
    idx_[key].push_back(i);
  }
}

std::vector<std::pair<int, double>> CubicIndex::nbs_with_dist(
    const Eigen::Vector3d &a,
    double distance, bool sorted) const {
  assert(distance > 0);
  const auto sq_dist_th = distance * distance;
  const auto potential_nbs = nb_candidates(a, distance);
  std::vector<std::pair<int, double>> nbs;
  for (auto i: potential_nbs) {
    const auto d = (*pxyz_)[i] - a;
    const double sq_dist = d.squaredNorm();
    if (sq_dist < sq_dist_th) { nbs.push_back(std::pair<int, double>(i, sq_dist)); }
  }
  if (sorted) {
    std::sort(nbs.begin(), nbs.end(), 
              [](std::pair<int, double> a, std::pair<int, double> b) {
                return a.second < b.second;
              });
  }
  return nbs;
}

std::vector<std::string> CubicIndex::nb_cells(const Eigen::Vector3d &a,
                                              double r) const {
  auto offset = ICEIL(r/grid_);
  Eigen::Vector3i ig;
  FOR(i, kDim) ig(i) = ICEIL(a(i)/grid_);
  std::vector<std::string> result;
  RANGE(i, ig(0)-offset, ig(0)+offset+1) {
    RANGE(j, ig(1)-offset, ig(1)+offset+1) {
      RANGE(k, ig(2)-offset, ig(2)+offset+1) {
        auto key = STR(i)+','+STR(j)+','+STR(k);
        if (idx_.count(key) > 0) { result.push_back(key); }
      }
    }
  }
  return result;
}

StdVecI CubicIndex::nb_candidates(const Eigen::Vector3d &a, double r) const {
  StdVecI result;
  const auto grids = nb_cells(a, r);
  for (auto k: grids) {
    result.insert(result.end(), idx_.at(k).begin(), idx_.at(k).end());
  }
  return result;
}

// Fixed-radius near neighbors
std::unordered_map<int, StdVecI> nbs_r(const CubicIndex &cache, double r) {
  const double sq_dist_th = r * r;
  std::unordered_map<int, StdVecI> result;
  std::unordered_set<std::string> processed;
  const auto xyz = *(cache.pxyz_);
  for (auto cell_i: cache.idx_) {
    const auto a_on_grid = *(cell_i.second.begin());
    const auto nb_grid_ks = cache.nb_cells(xyz[a_on_grid], r);
    for (auto key_j: nb_grid_ks) {
      if (processed.count(key_j) > 0 ) { continue; }
      for (auto i: cell_i.second) {
        for (auto j: cache.idx_.at(key_j)) {
          const auto d = xyz[j] - xyz[i];
          const double sq_dist = d.squaredNorm();
          if (sq_dist < sq_dist_th) {
            result[i].push_back(j);
            result[j].push_back(i);
          }
        }
      }
    }
    processed.insert(cell_i.first);
  }
  return result;
}


StdVecI CubicIndex::query(const Eigen::Vector3d &a, double distance,
                          bool sorted) const {
  auto nbs =  nbs_with_dist(a, distance, sorted);
  StdVecI result(nbs.size());
  FOR(i, nbs.size()) {result[i] = nbs[i].first; }
  return result;
}

// Grid initialize and filter
StdVecPt InitGrids(const StdVecPt &xyz, const StdVecD &r, double grid_size) {
  StdVecPt grids;
  const double r_max = *std::max_element(r.begin(), r.end());
  // Define the space: O, NumGrids
  Eigen::Vector3d o(FLT_MAX, FLT_MAX, FLT_MAX);
  Eigen::Vector3d o_max(FLT_MIN, FLT_MIN, FLT_MIN);
  Eigen::Vector3i num_grids;
  for (auto a: xyz) {
    FOR(i, kDim) {
      o(i) = std::min(o(i), a(i));
      o_max(i) = std::max(o_max(i), a(i));
    }
  }
  FOR(i, kDim) { 
    o(i) = long(floor((o(i) - r_max)/grid_size)) * grid_size;
    num_grids(i) = int((o_max(i) + r_max - o(i))/grid_size + 1.0);
  }
  // Span and filter
  const CubicIndex frame(xyz, r_max);
  FOR(i, num_grids(0)) {
    double x = o(0) + i * grid_size;
    FOR(j, num_grids(1)) {
      double y = o(1) + j * grid_size;
      FOR(k, num_grids(2)) {
        // if ((i%2 + j%2 + k%2) % 2 != 0) { continue; } // Ignore none FCC Point
        double z = o(2) + k * grid_size;
        Eigen::Vector3d g(x, y, z);
        const StdVecI nbs = frame.query(g, r_max);
        for (auto n: nbs) {
          Eigen::Vector3d d = xyz[n] - g;
          if (d.squaredNorm() < r[n] * r[n]) {
            grids.push_back(g);
            break;
          }
        }
      }
    }
  }
  return grids;
}

StdVecPt grid_filter(const StdVecPt &grids, const StdVecPt &points,
                     const StdVecD &r,
                     GFMod mod) {
  StdVecPt filtered;
  const double r_max = *std::max_element(r.begin(), r.end());
  const CubicIndex frame(points, r_max);
  FOR(i, grids.size()) {
    bool has_it = ((mod == kNotWithin) ? true : false);
    const StdVecI nbs = frame.query(grids[i], r_max);
    for (auto n: nbs) {
      Eigen::Vector3d d = grids[i] - points[n];
      if (d.squaredNorm() < r[n] * r[n])  {
        has_it = !has_it;
        break;
      }
    }
    if (has_it) filtered.push_back(grids[i]);
  }
  return filtered;
}

void mask(const Eigen::Vector3d &xyza, double ra,
          const Eigen::Vector3d &xyzb, double rb,
          const Eigen::MatrixXd &sample_dots,
          StdVecB *maskbit) {
  const double threshold_l = 0.0001;
  const double threshold_u = ra + rb;
  Eigen::Vector3d dir = xyzb - xyza;
  const double d = dir.norm();
  if (d > threshold_l and d < threshold_u) { // Not check itself and not-contact pairs
    const double r_proj = (ra * ra + d * d - rb * rb) / (2.0 * d);
    dir = dir.array() / d;
    Eigen::VectorXd smpl_proj = sample_dots * dir;
    smpl_proj = smpl_proj.array() * ra;
    FOR(i, maskbit->size()) {
      if (smpl_proj[i] > r_proj) (*maskbit)[i] = false;
    }
  }
}

//
std::vector<StdVecB> asa_masks(
    const StdVecPt &xyz, const Eigen::VectorXd &r,
    const Eigen::MatrixXd &smpl_dots) {
    const double r_max = r.maxCoeff();
    const double grid = 2 * r_max;
    const bool sorted = true;
    const CubicIndex frame(xyz, grid);
    std::vector<StdVecB> masks(xyz.size(), StdVecB(smpl_dots.rows(), true));
    FOR(i, xyz.size()) {
      const StdVecI nbs = frame.query(xyz[i], grid, sorted);
      for (auto j : nbs) {
        if (j == i)  continue;  // The point itself might be returned by query
        mask(xyz[i], r(i), xyz[j], r(j), smpl_dots, &(masks[i]));
        int sum = std::count(masks[i].begin(), masks[i].end(), true);
        if (sum == 0) break;  // Next if all sampling dots are covered
      }
    }
    return masks;
}

StdVecD nn_dist(const StdVecPt &xyzs, const StdVecPt &dots, double dist) {
  const double grid_space = 5.0;
  StdVecD nearest_neighbor_distance(xyzs.size(), dist);
  const CubicIndex frame(xyzs, grid_space);
  FOR(i, dots.size()) {
    const StdVecI nbs = frame.query(dots[i], dist);
    for (auto j: nbs) {
      Eigen::Vector3d d = dots[i] - xyzs[j];
      double dist = d.norm();
      if (nearest_neighbor_distance[j] > dist) {
        nearest_neighbor_distance[j] = dist;
      }
    }
  }
  return nearest_neighbor_distance;
}

void vec_show_as_string(const StdVecI &v) {
  std::string result;
  for (auto x: v) { result += STR(x) + ';'; }
  std::cout << result << std::endl;
}

StdVecI dbscan(const StdVecPt &dots,
               double eps, int min_points) {
  const int noise_class = 0;
  const CubicIndex indexed(dots, eps);
  const auto nbs = nbs_r(indexed, eps);
  // std::cout << "#" << dots.size() << " " << nbs.size() << std::endl;
  StdVecI classids(dots.size(), noise_class);
  StdVecB processed(dots.size(), false);
  int cnt = 0;
  FOR(i, dots.size()) {
    if (processed[i]) continue;
    processed[i] = true;
    StdVecI seeds = nbs.at(i);
    if (seeds.size()-1 >= min_points) {  // -1: remove dots[i] from the count
      classids[i] = ++cnt;  // Create cluster
      std::unordered_set<int> uniq_seeds(seeds.begin(), seeds.end());
      int pos = 0;
      while (pos < seeds.size()) { // Expand cluster
        int i_nb = seeds[pos];
        if (! processed[i_nb]) {  // dots[i] is skipped
          processed[i_nb] = true;
          const auto current_nbs = nbs.at(i_nb);
          // Extend neighbor list
          if (current_nbs.size()-1 >= min_points) {
            for (auto j: current_nbs) {
              if (uniq_seeds.find(j) == uniq_seeds.end()) {
                uniq_seeds.insert(j);
                seeds.push_back(j);
              } 
            }
          }
          if (classids[i_nb] == noise_class) {
            classids[i_nb] = cnt;
          }
        }
        ++pos;
      }
    }
  }
  return classids;
}

void dot2pdb(const StdVecPt &dots, const char *filename) {
  const std::vector<char> chain_ids(dots.size(), '-');
  StdVecI res_ids(dots.size(), 0);
  FOR(i, dots.size()) { res_ids[i] = int(i/1000); }
  dot2pdb(dots, filename, res_ids, chain_ids);
}

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecI &res_ids) {
  const std::vector<char> chain_ids(dots.size(), '-');
  dot2pdb(dots, filename, res_ids, chain_ids);
}

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecC &chain_ids) {
  StdVecI res_ids(dots.size(), 0);
  FOR(i, dots.size()) { res_ids[i] = int(i/1000); }
  dot2pdb(dots, filename, res_ids, chain_ids);
}

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecI &res_ids,
             const std::vector<char> &chain_ids) {
  FILE *outfile;
  outfile = fopen(filename, "w");
  FOR(i, dots.size()) {
    fprintf(outfile,
             "ATOM  %5i  %-3s %3s %c%4i    %8.3f%8.3f%8.3f      %6.2f\n",
            int(i%100000), "C", "UNK", chain_ids[i], res_ids[i]%10000,
             dots[i](0), dots[i](1), dots[i](2), 0.0);
  }
  fclose(outfile);
}

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecD &bfactors) {
  const std::vector<char> chain_ids(dots.size(), '-');
  StdVecI res_ids(dots.size(), 0);
  FOR(i, dots.size()) { res_ids[i] = int(i/1000); }
  //
  FILE *outfile;
  outfile = fopen(filename, "w");
  FOR(i, dots.size()) {
    fprintf(outfile,
             "ATOM  %5i  %-3s %3s %c%4i    %8.3f%8.3f%8.3f      %6.2f\n",
            int(i%100000), "C", "UNK", chain_ids[i], res_ids[i]%10000,
             dots[i](0), dots[i](1), dots[i](2), bfactors[i]);
  }
  fclose(outfile);
}

std::vector<StdVecPt> mask2dot(
    const StdVecPt &xyz, const Eigen::VectorXd &r,
    const std::vector<StdVecB> &masks,
    const Eigen::MatrixXd &smpl_dots) {
  std::vector<StdVecPt> dot_sets;
  FOR(i, xyz.size()) {
    Eigen::MatrixXd surf_dot = smpl_dots.array() * r(i);
    surf_dot.rowwise() += xyz[i].transpose();
    dot_sets.push_back(StdVecPt());
    FOR(j, smpl_dots.rows()) {
      if (masks[i][j]) (*dot_sets.rbegin()).push_back(surf_dot.row(j));
    }
  }
  return dot_sets;
}

StdVecPt get_largest_cluster(const StdVecPt &asa_points, const StdVecI &cluster_id) {
  // Count cluster
  std::unordered_map<int, int> count;
  for (auto i: cluster_id) {
    if (count.find(i) == count.end()) {
      count[i] = 1;
    } else {
      count[i] += 1;
    }
  }
  // Get the largest cluster (the real ASA surface most likely to be the largest one)
  int largest_cluster_id = -1;
  int largest_cnt = -1;
  for (const auto &n: count) {
    if (n.second > largest_cnt) {
      largest_cluster_id = n.first;
      largest_cnt = n.second;
    }
  }
  StdVecPt largest_cluster;
  FOR(i, asa_points.size()) {
    if (cluster_id[i] == largest_cluster_id) {
      largest_cluster.push_back(asa_points[i]);
    }
  }
  return largest_cluster;
}

// Apps
StdVecD depth(const std::vector<StdVecD> &xyzs,
              const StdVecD &rs,
              const std::vector<StdVecD> &dots,
              double probe,
              double distance,
              const StdVecD &plane_o,
              const StdVecD &plane_n,
              bool remove_interior_cavities) {
  // Validation of arguments
  assert(xyzs.size() > 0);
  assert(rs.size() > 0);
  assert(dots.size() > 0);
  assert(xyzs.size() == rs.size());
  assert(xyzs[0].size() == kDim);
  assert(dots[0].size() == kDim);
  // Convert input arguments to Eigen
  StdVecPt xyz(xyzs.size());
  Eigen::VectorXd r(rs.size());
  FOR(i, xyzs.size()) {
    xyz[i] = Eigen::Vector3d(xyzs[i].data());
    r(i) = rs[i] + probe;
  }
  Eigen::MatrixXd smpl_dots(dots.size(), dots[0].size());
  FOR(i, dots.size()) {
    FOR(j, dots[i].size()) {
      smpl_dots(i, j) = dots[i][j];
    }
  }
  // Get ASA surface
  const std::vector<StdVecB> masks = asa_masks(xyz, r, smpl_dots);
  const std::vector<StdVecPt> dot_sets = mask2dot(xyz, r, masks, smpl_dots);
  StdVecPt asa_points;
  //x StdVecI res_ids;
  //x int i = 0;
  if (plane_o.size() == 0) {
    for (auto dots: dot_sets) {
      asa_points += dots;
      //x for (auto dot: dots) { res_ids.push_back(i); }
      //x ++i;
    }
  } else {
    // Filter out ASA surface by a plane 
    Eigen::Vector3d o(plane_o.data());
    Eigen::Vector3d n(plane_n.data());
    for (auto dots: dot_sets) { 
      for (auto dot: dots) {
        Eigen::Vector3d dir = dot - o;
        if (dir.dot(n) > 0.0){  // Filter all plane masked cases. In HIV-Env,
                              // the plane is located at membran+probe_radius
          asa_points.push_back(dot);
          //x res_ids.push_back(i);
        }
      }
      //x ++i;
    }
  }
  //x dot2pdb(asa_points, "out_cpp.pdb", res_ids);
  if (not remove_interior_cavities) {
    //std::cout<< "No remove_interior_cavities" << std::endl;
    return nn_dist(xyz, asa_points, probe + distance);
  } else {
    //std::cout<< "Remove_interior_cavities" << std::endl;
    double min_cluster_size = 4;
    double max_r = *std::max_element(rs.begin(), rs.end());
    double eps = get_eps(probe+max_r, dots.size());
    const StdVecI cluster_id = dbscan(asa_points, eps, min_cluster_size);
    //std::cout << "#" << asa_points.size() << std::endl;
    StdVecPt asa_points_top_cluster = get_largest_cluster(asa_points, cluster_id);
    return nn_dist(xyz, asa_points_top_cluster, probe + distance);
  }
}

// Depth: seperate points for surface definition (Calpha atoms) and depth calculation (all atoms)
StdVecD depth2(const std::vector<StdVecD> &surf_xyzs,
               const std::vector<StdVecD> &depth_xyzs,
               const std::vector<StdVecD> &dots,
               double probe,
               double distance) {
  // Validation of arguments
  assert(surf_xyzs.size() > 0);
  assert(depth_xyzs.size() > 0);
  assert(dots.size() > 0);
  assert(surf_xyzs[0].size() == kDim);
  assert(depth_xyzs[0].size() == kDim);
  assert(dots[0].size() == kDim);
  // Convert input arguments to Eigen
  StdVecPt surf_xyz(surf_xyzs.size());
  Eigen::VectorXd r(surf_xyzs.size());
  FOR(i, surf_xyzs.size()) {
    surf_xyz[i] = Eigen::Vector3d(surf_xyzs[i].data());
    r(i) = probe;
  }
  StdVecPt depth_xyz(depth_xyzs.size());
  FOR(i, depth_xyzs.size()) {
    depth_xyz[i] = Eigen::Vector3d(depth_xyzs[i].data());
  }
  Eigen::MatrixXd smpl_dots(dots.size(), dots[0].size());
  FOR(i, dots.size()) {
    FOR(j, dots[i].size()) {
      smpl_dots(i, j) = dots[i][j];
    }
  }
  // Get ASA surface
  const std::vector<StdVecB> masks = asa_masks(surf_xyz, r, smpl_dots);
  const std::vector<StdVecPt> dot_sets = mask2dot(surf_xyz, r, masks, smpl_dots);
  StdVecPt asa_points;
  //x StdVecI res_ids;
  //x int i = 0;
  for (auto dots: dot_sets) {
    asa_points += dots;
    //x for (auto dot: dots) { res_ids.push_back(i); }
    //x ++i;
  }
  //x dot2pdb(asa_points, "out_cpp1.pdb", res_ids);

  double min_cluster_size = 4;
  double eps = get_eps(probe, dots.size());
  const StdVecI cluster_id = dbscan(asa_points, eps, min_cluster_size);
  StdVecPt asa_points_top_cluster = get_largest_cluster(asa_points, cluster_id);
  //x dot2pdb(asa_points, "out_cppx.pdb", cluster_id);

  //x Eigen::VectorXd rx(asa_points_top_cluster.size());
  //x FOR(i, asa_points_top_cluster.size()) { rx(i) = probe; }
  //x const std::vector<StdVecB> masksx = asa_masks(asa_points_top_cluster, rx, smpl_dots);
  //x const std::vector<StdVecPt> dot_setsx = mask2dot(asa_points_top_cluster, rx, masksx, smpl_dots);
  //x StdVecPt asa_pointsx;
  //x for (auto dots: dot_setsx) { asa_pointsx += dots; }
  //x const StdVecI cluster_idx = dbscan(asa_pointsx, eps, min_cluster_size);
  //x dot2pdb(asa_pointsx, "out_cpp2.pdb", cluster_idx);

  return nn_dist(depth_xyz, asa_points_top_cluster, probe + distance);
}

double get_eps(double probe, int n_dots) {
  double pi = 3.14159;
  double ideal_area = 4.0/3.0 * pi * probe * probe / n_dots;
  double ideal_dist = sqrt(ideal_area * 4.0 / sqrt(3.0));
  double epsilon = 1.5 * ideal_dist;
  return epsilon;
}

StdVecD asa_dots(const std::vector<StdVecD> &xyzs,
                 const StdVecD &rs,
                 const std::vector<StdVecD> &dots,
                 double probe,
                 bool remove_interior_cavities) {
  // Validation of arguments
  assert(xyzs.size() > 0);
  assert(rs.size() > 0);
  assert(dots.size() > 0);
  assert(xyzs.size() == rs.size());
  assert(xyzs[0].size() == kDim);
  assert(dots[0].size() == kDim);
  // Convert input arguments to Eigen
  StdVecPt xyz(xyzs.size());
  Eigen::VectorXd r(rs.size());
  FOR(i, xyzs.size()) {
    xyz[i] = Eigen::Vector3d(xyzs[i].data());
    r(i) = rs[i] + probe;
  }
  Eigen::MatrixXd smpl_dots(dots.size(), dots[0].size());
  FOR(i, dots.size()) {
    FOR(j, dots[i].size()) {
      smpl_dots(i, j) = dots[i][j];
    }
  }
  // Get ASA surface
  const std::vector<StdVecB> masks = asa_masks(xyz, r, smpl_dots);
  const std::vector<StdVecPt> dot_sets = mask2dot(xyz, r, masks, smpl_dots);
  // Count ASA 
  StdVecD accessible_area(dot_sets.size());
  const int num_smpl = smpl_dots.rows();
  if (not remove_interior_cavities) {
    FOR (i, dot_sets.size()) {
      int n_surface = dot_sets[i].size();
      const double asa_i = 4.0 * 3.14159 * r(i) * r(i) * n_surface / num_smpl;
      accessible_area[i] = asa_i;
    }
  } else {
    // Extract the largest dot cluster 
    StdVecPt asa_points;
    for (auto dots: dot_sets) { asa_points += dots; }
    int min_cluster_size = 4;
    double eps = get_eps(probe, dots.size());
    const StdVecI cluster_id = dbscan(asa_points, eps, min_cluster_size);
    StdVecPt asa_extracted = get_largest_cluster(asa_points, cluster_id);
    // Count only dots in the largest dot cluster (thus, interior cavities are removed)
    const CubicIndex asa_surface(asa_extracted, 0.01);
    FOR (i, dot_sets.size()) {
      int n_surface = 0;
      for (auto dot: dot_sets[i]) {
        const StdVecI nbs = asa_surface.query(dot, 0.01);  // Estimtaed distance between dots ~0.3
        if (nbs.size() > 0) { ++n_surface; }
      }
      const double asa_i = 4.0 * 3.14159 * r(i) * r(i) * n_surface / num_smpl;
      accessible_area[i] = asa_i;
    }
  }
  return accessible_area;
}
