#include <string>
#include <vector>
#include <algorithm>
#include <unordered_map>
#include <unordered_set>

#include <cfloat>
#include <cmath>
#include <cassert>

#include <Eigen/Dense>

// C++11 way
#define DISALLOW_COPY_AND_ASSIGN(TName) \
TName(TName&) = delete;                 \
void operator=(TName) = delete;

#define FOR(i, n)  for (long M_TMP = (n), (i) = 0; (i) < M_TMP; ++(i))
#define RANGE(i, a, b) for (long M_TMP = (b), (i) = (a); (i) < M_TMP; ++(i))

#define ICEIL(x) static_cast<long>(ceil(x))
#define STR(x) std::to_string(x)

static const int kDim =  3;

typedef std::vector<int> StdVecI;
typedef std::vector<bool> StdVecB;
typedef std::vector<char> StdVecC;
typedef std::vector<double> StdVecD;
typedef std::vector<Eigen::Vector3d> StdVecPt;

// Concatenate vectors
template <typename T>
std::vector<T> operator+(const std::vector<T> &a, const std::vector<T> &b) {
  std::vector<T> ab;
  ab.reserve(a.size() + b.size());
  ab.insert(ab.end(), a.begin(), a.end());
  ab.insert(ab.end(), b.begin(), b.end());
  return ab;
}

template <typename T>
std::vector<T> &operator+=(std::vector<T> &a, const std::vector<T> &b) {
  a.reserve(a.size() + b.size());
  a.insert(a.end(), b.begin(), b.end());
  return a;
}

// Cubic spacial index
class CubicIndex {
 public:
  CubicIndex(const StdVecPt &xyz, double index_grid);
  StdVecI query(const Eigen::Vector3d &a, double distance, 
                bool sorted=false) const;
  std::vector<std::pair<int, double>> nbs_with_dist(
      const Eigen::Vector3d &a,
      double distance,
      bool sorted=false) const;
  friend std::unordered_map<int, StdVecI> nbs_r(const CubicIndex &frame, double r);
 private:
  std::string get_key(const Eigen::Vector3d &a) const;
  std::vector<std::string> nb_cells(const Eigen::Vector3d &a, double r) const;
  StdVecI nb_candidates(const Eigen::Vector3d &a, double r) const;

  const StdVecPt *pxyz_;
  double grid_;
  std::unordered_map<std::string, StdVecI> idx_;
  // DISALLOW_COPY_AND_ASSIGN(CubicIndex);
};

StdVecPt init_grids(const StdVecPt &xyz, const StdVecD &r, double grid_size);

enum GFMod {kNotWithin=0, kWithin};

StdVecPt grid_filter(
    const StdVecPt &grids, const StdVecPt &points,
    const StdVecD &r, GFMod mod);

void mask(const Eigen::Vector3d &xyza, double ra,
          const Eigen::Vector3d &xyzb, double rb,
          const Eigen::MatrixXd &sample_dots,
          StdVecB *maskbit);

std::vector<StdVecB> asa_masks(
    const StdVecPt &xyz, const Eigen::VectorXd &r,
    const Eigen::MatrixXd &smpl_dots);

// Output dots out in pdb format
void dot2pdb(const StdVecPt &dots, const char *filename);

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecI &res_ids);

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecC &chain_ids);

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecI &res_ids,
             const StdVecC &chain_ids);

void dot2pdb(const StdVecPt &dots, const char *filename,
             const StdVecD &bfactors);

// Cluster by DBSCAN. Ref: http://en.wikipedia.org/wiki/DBSCAN
StdVecI dbscan(const std::vector<Eigen::Vector3d > &dots,
            double eps, int min_points);

double get_eps(double probe, int n_dots);

StdVecD nn_dist(const StdVecPt &xyzs, const StdVecPt &dots, double dist);

StdVecD depth(const std::vector<StdVecD> &xyzs,
              const StdVecD &rs,
              const std::vector<StdVecD> &dots,
              double probe,
              double distance,
              const StdVecD &plane_o,
              const StdVecD &plane_n,
              bool remove_interior_cavities=false);

StdVecD depth2(const std::vector<StdVecD> &surf_xyzs,
               const std::vector<StdVecD> &depth_xyzs,
               const std::vector<StdVecD> &dots,
               double probe,
               double distance);

StdVecPt get_largest_cluster(const StdVecPt &asa_points, const StdVecI &cluster_id);

StdVecD asa_dots(const std::vector<StdVecD> &xyzs,
                const StdVecD &rs,
                const std::vector<StdVecD> &dots,
                double probe,
                bool remove_interior_cavities=false);
