# Collection of constants and utilities
import sys
import os
import math
import numpy
import argparse
import collections
import itertools as it

iterable = lambda obj: isinstance(obj, collections.Iterable)
first = lambda iterable_obj: next(iter(iterable_obj))

EntLevels='SMCRA'  # Bio.PDB organizes protein structure as SMCRA (Structure-Model-Chain-Residue-Atom)
BackBoneAtoms = set('C N O CA'.split())

def cleaned(instream, comment='#'):
    """ Remove comments(starts with #) or empty lines """
    for line in instream:
        cleaned_line = line.strip().split(comment)[0]
        if cleaned_line:
            yield cleaned_line.strip()

Glycans = set('NAG MAN BMA FUC BGC NDG FUL GAL SIA'.split())
AAs = set('ALA CYS ASP GLU PHE GLY HIS ILE LYS LEU MET ASN PRO GLN ARG SER THR VAL TRP TYR'.split())
AAs.update('MSE HSE HSD'.split())
AAchar = 'GASTCVILMPFYWHKRDENQ'
AA3to1 = {'ALA': 'A', 'CYS': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F',
          'GLY': 'G', 'HIS': 'H', 'ILE': 'I', 'LYS': 'K', 'LEU': 'L',
          'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R',
          'SER': 'S', 'THR': 'T', 'VAL': 'V', 'TRP': 'W', 'TYR': 'Y',
          'MSE': 'M', 'UNK': 'X'}

# AA size: number of heavy atoms in the side-chain + CA
AASize = {'G':1, 'A':2, 'C':3, 'S':3, 'T':4, 'V':4, 'P':4, 'I':5, 'L':5, 'M':5,
          'N':5, 'D':5, 'B': 5, 'E':6, 'Q':6, 'Z': 6,
          'K':6, 'H':7, 'R':8, 'F':8, 'Y':9, 'W':11}

# 20 colors (https://sashamaps.net/docs/resources/20-colors/)
color20 = ['#800000', '#9A6324', '#808000', '#469990', '#000075',
           '#e6194B', '#f58231', '#ffe119', '#bfef45', '#3cb44b',
           '#42d4f4', '#4363d8', '#911eb4', '#f032e6', '#fabed4',
           '#ffd8b1', '#fffac8', '#aaffc3', '#dcbeff', '#a9a9a9']
color20names = [ 'Maroon', 'Brown', 'Olive', 'Teal', 'Navy',
                 'Red', 'Orange', 'Yellow', 'Lime', 'Green',
                 'Cyan', 'Blue', 'Purple', 'Magenta', 'Pink',
                 'Apricot', 'Beige', 'Mint', 'Lavender', 'Grey']

_aa_na_correspondence = '''
    # Amino_acid	Codons	Compressed_Codons
    Ala/A	GCT,GCC,GCA,GCG	GCN
    Arg/R	CGT,CGC,CGA,CGG,AGA,AGG	CGN,MGR
    Asn/N	AAT,AAC	AAY
    Asp/D	GAT,GAC	GAY
    Cys/C	TGT,TGC	TGY
    Gln/Q	CAA,CAG	CAR
    Glu/E	GAA,GAG	GAR
    Gly/G	GGT,GGC,GGA,GGG	GGN
    His/H	CAT,CAC	CAY
    Ile/I	ATT,ATC,ATA	ATH
    Leu/L	TTA,TTG,CTT,CTC,CTA,CTG	YTR,CTN
    Lys/K	AAA,AAG	AAR
    Met/M	ATG
    Phe/F	TTT,TTC	TTY
    Pro/P	CCT,CCC,CCA,CCG	CCN
    Ser/S	TCT,TCC,TCA,TCG,AGT,AGC	TCN,AGY
    Thr/T	ACT,ACC,ACA,ACG	ACN
    Trp/W	TGG
    Tyr/Y	TAT,TAC	TAY
    Val/V	GTT,GTC,GTA,GTG	GTN
    STOP/*	TAA,TGA,TAG	TAR,TRA
    #START	ATG

    # From inverse table of "http://en.wikipedia.org/wiki/DNA_codon_table". Converted by
    #> gawk -v FS="[,\t /]" -v format="'%s':'%s'," '{if ($0!~/^#/){for (i=3;i<=NF;i++) {printf(format, $i, $2) } print();}}' aa2na.txt
'''
NA2AA = {}
for line in cleaned(_aa_na_correspondence.split('\n')):
    fields = line.split('\t')
    aa3char, aa1char = fields[0].split('/')
    NA2AA.update((codon, aa1char) for codons in fields[1:] for codon in codons.split(','))
CompressedCodon = set('GCN CGN MGR AAY GAY TGY CAR GAR GGN CAY ATH YTR CTN AAR TTY CCN TCN AGY ACN TAY GTN TAR TRA'.split())

# source of total area: https://en.wikipedia.org/wiki/Relative_accessible_surface_area
# Rost, Burkhard; Sander, Chris (November 1994). Proteins. 20. PMID 7892171
# "Conservation and Prediction of Solvent Accessibility in Protein Families".
#
# B stands for D or N; Z for E or Q and X for an undetermined Amino acid
TotalASA = {'A': 106, 'C': 135, 'D': 163, 'E': 194, 'F': 197,
            'G':  84, 'H': 184, 'I': 169, 'K': 205, 'L': 164,
            'M': 188, 'N': 157, 'P': 136, 'Q': 198, 'R': 248,
            'S': 130, 'T': 142, 'V': 142, 'Y': 222, 'W': 227,
            'B': 160, 'X': 180, 'Z': 196}

# PDB ID < 50% subset, 20230127
AAFreq = {'L': 0.0945, 'A': 0.0794, 'G': 0.0695, 'E': 0.0680, 'V': 0.0671,
          'S': 0.0658, 'K': 0.0593, 'D': 0.0570, 'I': 0.0557, 'T': 0.0539,
          'R': 0.0523, 'P': 0.0464, 'N': 0.0435, 'F': 0.0396, 'Q': 0.0389,
          'Y': 0.0332, 'H': 0.0271, 'M': 0.0227, 'C': 0.0132, 'W': 0.0129}

BondThreshold = 2.0

# Substitution Matrix: BLOSUM62, Benner74, VTML200, MIQS
from Bio.Align import substitution_matrices as submat
Benner74 = submat.load('Benner74')
Benner74 = dict(((a, b), Benner74[a, b]) for (a,b) in it.product(Benner74.alphabet, Benner74.alphabet))
Blosum62 = submat.load('BLOSUM62')
Blosum62 = dict(((a, b), Blosum62[a, b]) for (a,b) in it.product(Blosum62.alphabet, Blosum62.alphabet))

def read_matrix(infilestream):
    """ Read VTML200-like matrix files (e.g. BLOSUM62) """
    content = [line.split() for line in cleaned(infilestream)]
    matrix = {}
    restype = content[0]
    for elements in content[1:]:
        restypeA = elements[0]
        for restypeB, score in zip(restype, elements[1:]):
            matrix[(restypeA, restypeB)] = float(score)
    return matrix

path_to_this_script = os.path.dirname(os.path.abspath(__file__))
with open(path_to_this_script+'/data/VTML200.txt') as infile:
    VTML200 = read_matrix(infile)

# https://academic.oup.com/peds/article-abstract/7/11/1323/1479698?redirectedFrom=fulltext
with open(path_to_this_script+'/data/MIQS.txt') as infile:
    MIQS = read_matrix(infile)

with open(path_to_this_script+'/data/HIVMat38.txt') as infile:
    HIVMat38 = read_matrix(infile)

class Plane:
    def __init__(self, A, B, C):
        assert(isinstance(A, numpy.ndarray))
        assert(isinstance(B, numpy.ndarray))
        assert(isinstance(C, numpy.ndarray))
        ab = B - A
        ac = C - A
        n = numpy.cross(ab, ac)
        self.n = n / numpy.linalg.norm(n)
        self.o = (A + B + C) / 3.0

    def dist(self, x):
        assert(isinstance(x, numpy.ndarray))
        return numpy.dot((x - self.o), self.n)

    def translation(self, d):
        self.o += self.n * d

# grid sasa sample points
smpl256  = numpy.loadtxt(path_to_this_script+'/' + 'data/256.txt' )[:,0:3]
smpl2048 = numpy.loadtxt(path_to_this_script+'/' + 'data/2048.txt')[:,0:3]

def print_hist(data, file, step=1.0):
    lower = math.floor(min(data) / step) * step - step
    upper = math.ceil(max(data)/ step) * step + step * 1.01
    bins = numpy.arange(lower, upper, step)
    hist, bins = numpy.histogram(data, bins=bins)
    print('# Frequency in each bin:', file=file)
    for i, f in enumerate(hist[:-1]):
        if f > 0:
            print('# [%.1f, %.1f):%4d' %(bins[i], bins[i+1], f), file=file)
    print('# [%.1f, %.1f]:%4d' %(bins[-2], bins[-1], hist[-1]), file=file)  # Last bin
    print('', file=file)

def zopen(fname, specifications='rt', msg=None):
    " open compressed files "
    if fname.endswith('.gz'):
        import gzip
        open_fun = gzip.open
    elif fname.endswith('.bz2'):
        import bz2
        open_fun = bz2.open
    else:
        open_fun = open
    try:
        return open_fun(fname, specifications)
    except IOError:
        print('File open error. %s cannot be opened' %(fname))
        if msg is not None: print(msg)
        sys.exit()

class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)

