import sys
import collections
import copy
import math
import subprocess
import tempfile
import shutil  # Tested: os cannot remove directory that are not empty
import re
import itertools as it

try:
    from . import misc
except Exception:
    import misc

def hamming_distance(s1, s2):
    """Return the Hamming distance between equal-length sequences"""
    if len(s1) != len(s2):
        raise ValueError("Undefined for sequences of unequal length")
    return sum(el1 != el2 for el1, el2 in zip(s1, s2))

def score_pairwise(seq1, seq2, matrix, gap_s=-5, gap_e=-1, gap=True, empty_col=0):
    """ usage: score = sum(s for s in score_pairwise(cons1, cons2, Blosum62, -5, -1)) """
    for A,B in zip(seq1, seq2):
        diag = ('-'==A) or ('-'==B)
        matched_gap = ('-' == A and '-' == B)
        yield (empty_col if matched_gap else (gap_e if gap else gap_s)) if diag else matrix[(A.upper(),B.upper())]
        gap = diag

def identity(seq1, seq2, gap_collapsed=True):
    matched_len = 0.0
    total_len = 0.0
    gapopen = False
    for ai, bi in zip(seq1, seq2):
        if ai == '-' and bi == '-': # empty column
            continue
        elif ai == '-' or bi == '-': # gap
            if gap_collapsed:  # consecutive gaps count as one difference
                if not gapopen:
                    gapopen = True
                    total_len +=1.0
            else:
                total_len +=1.0
        elif ai == bi: # matched
            matched_len += 1.0
            total_len += 1.0
            gapopen = False
        else: # mismatched, non-gap
            total_len += 1.0
            gapopen = False
    return matched_len/total_len


def w_sim(refseq, seq, weights, matrix=misc.Blosum62):
    """  Weighted Blosum62 similiarity score between seqa and seqb

    seqa, seqb: two sequences to be compared of
    weights: weight dictionary for sites that has a weight > 0 ({i: w_i; j: w_j})
    """
    is_indel = lambda char: char in '.-#*'  # Insertion/Deletion/Frameshift/Stop
    indel_score = min(matrix.values())
    score = 0
    for i, w in weights.items():
        r = refseq[i]
        a = seq[i]
        if is_indel(r) and is_indel(a):
            continue
        elif is_indel(r) or is_indel(a):
            score += indel_score * w
        else:
            score += matrix[(r, a)] * w
    return score/sum(weights.values())

def w_dist(refseq, seq, weights, matrix=misc.Blosum62, symmetric=True):
    """  Similiarity of seq to refseq

    refseq: the reference sequence
    seq: the sequcne to be analyzed
    weights: weight dictionary ({i: w_i; ...})
    """
    is_indel = lambda char: char in '.-#*'  # Insertion/Deletion/Frameshift/Stop
    indel_score = min(matrix.values())
    score = 0
    for i, w in weights.items():
        r = refseq[i]
        a = seq[i]
        if is_indel(r) and is_indel(a):
            ref_score = 0
            match_score = 0
        elif is_indel(r) and (not is_indel(a)):
            ref_score = matrix[(a, a)]
            match_score = indel_score
        elif (not is_indel(r)) and is_indel(a):
            ref_score = matrix[(r, r)]
            match_score = indel_score
        else:
            if symmetric:
                ref_score = (matrix[(r, r)] + matrix[(a, a)])/2
            else:
                ref_score = matrix[(r, r)]
            match_score = matrix[(r, a)]
        score += w * (ref_score - match_score)
    return score/sum(weights.values())

def bestmatch(query_seq_aligned, msa):
    max_score = 0
    idx = 0
    for i, ref_seq in enumerate(msa):
        current_score = sum(s for s in score_pairwise(ref_seq, query_seq_aligned, misc.Blosum62, gap_s=0, gap_e=0))
        if max_score < current_score:
            max_score = current_score
            idx = i
    return idx

def conservation_score(msa, matrix=misc.Blosum62, gap_penalty=-5):
    return similarity(msa, matrix, gap_penalty)

def similarity(msa, matrix=misc.Blosum62, gap_penalty=-5):
    assert(type(msa) == list)
    assert(len(msa) > 1)
    aln_len = len(msa[0])
    scores = []
    for i, seqa in enumerate(msa):
        for seqb in msa[i+1:]:
            scores.append(tuple(s for s in score_pairwise(seqa, seqb, matrix, gap_s=gap_penalty, gap_e=gap_penalty, empty_col=gap_penalty)))
    num_pairs = len(scores)
    result = []
    for i in range(aln_len):
        result.append(sum(s[i] for s in scores)/float(num_pairs))
    return result

def mafft(seqs, protocol):
    prog = protocol
    temp = tempfile.NamedTemporaryFile(delete=False, mode='wt')
    for i, s in enumerate(seqs):
        write_fasta(temp, i, s)
    temp.flush()  # temp.seek(0) if the file handle is called by some python code
    ps = subprocess.Popen((prog, '--anysymbol', '--thread', '10', '--quiet', temp.name),
                          stdout=subprocess.PIPE)
    temp.close()
    ids, aligned_seqs = read_fasta(ps.stdout)
    return tuple(aligned_seqs)

def mafft_linsi(seqs):
    return mafft(seqs, protocol='mafft-linsi')

def mafft_add(existing_msa, seqs, other_args=None):
    mafft = 'mafft'
    tmpdir = tempfile.mkdtemp()  # temp dir for files
    msa_file = tmpdir + '/msa.fa'
    tobe_added = tmpdir + '/tobe_added.fa'
    with open(msa_file, 'wt') as msa_fs, open(tobe_added, 'wt') as tobe_added_fs:
        for i, s in enumerate(existing_msa):
            write_fasta(msa_fs, i, s)
        for i, s in enumerate(seqs):
            write_fasta(tobe_added_fs, i, s)
    cmd = [mafft, '--anysymbol', '--thread', '4', '--quiet']
    if other_args:
        assert(type(other_args) == list)
        cmd += other_args
    cmd += ['--add', tobe_added, msa_file]
    ps = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    ids, aligned_seqs = read_fasta(ps.stdout)
    shutil.rmtree(tmpdir)
    return tuple(aligned_seqs)

def needle(seq_a, seq_b, gap_s=10.0, gap_e=0.5):
    tmpdir = tempfile.mkdtemp()  # temp dir for files
    name_a = tmpdir+'/seq_a'
    name_b = tmpdir+'/seq_b'
    aligned = tmpdir+'/aligned'
    #aligned = 'aligned.aln'
    prog = 'needle'
    with open(name_a, 'wt') as outfile_a, open(name_b, 'wt') as outfile_b:
        write_fasta(outfile_a, name_a, seq_a)
        write_fasta(outfile_b, name_b, seq_b)
    stdout = subprocess.check_output(
            (prog, name_a, name_b, '-gapopen', str(gap_s), '-gapextend', str(gap_e),
                '-outfile', aligned, '-aformat3', 'fasta'),
            stderr=subprocess.STDOUT)
    with open(aligned) as instream:
        heads, aligned_seqs = read_fasta(instream)
    shutil.rmtree(tmpdir)
    return tuple(aligned_seqs)

def read_nexus(filestream):
    seq_info_loading = False 
    is_matrix = False
    for line in filestream.readlines():
        str = line.strip()
        elements = str.split()
        if is_matrix and len(elements)==2 :
            h, s = elements
            headers.append(h)
            sequences.append(s)
        elif is_matrix and str == 'END;':
            break
        elif str == 'MATRIX' and seq_info_loading:
            is_matrix = True
        elif str == 'BEGIN DATA;':
            headers = []
            sequences = []
            seq_info_loading = True
        else:
            pass
    return (headers, sequences)

def fasta_iter(filestream):
    faiter = (x[1] for x in it.groupby(filestream, lambda line: line[0] == ">"))
    for header in faiter:
        header_str = header.__next__()[1:].strip()
        seq = "".join(s.strip() for s in faiter.__next__())
        yield header_str, seq

def read_fasta(filestream):
    headers = []
    sequences = []
    for h, s in fasta_iter(filestream):
        headers.append(h)
        sequences.append(s)
    #for line in filestream.readlines():
    #    if type(line) is bytes:
    #        s = line.strip().decode()
    #    else:
    #        s = line.strip()
    #    if s.startswith('>'):
    #        headers.append(s[1:])
    #        sequences.append([])
    #    else:
    #        sequences[-1].append(s)
    #sequences = [''.join(seqs).replace(' ', '') for seqs in sequences]
    return (headers, sequences)

def chunkstring(string, length):
    return (string[0+i:length+i] for i in range(0, len(string), length))

def write_fasta(filestream, head, seq, length=60):
    filestream.write('>{}\n'.format(head))  # filestream must be text mode ('wt')
    for chunk in chunkstring(seq, length):
        filestream.write('{}\n'.format(chunk))

def idmap_msa2ref(msa, ref, not_aa='-*#'):  # MSA: a list of sequences
    assert(type(msa) is list)
    assert(len(msa) > 0)
    assert(len(msa[0]) == len(ref))  # msa and ref should be aligned
    codes='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    msa_t = transpose(msa)
    num_taxon = len(msa_t[0])
    idmap = {}  # {msa_id: ref_id}
    ref_i = 0
    ins_i = -1  # insertion ID, -1: not insertion, 0-51: insertion
    for msa_i, (ref_res, msa_res) in enumerate(zip(ref, msa_t)):
        cnt_not_aa = sum(msa_res.count(c) for c in not_aa)
        not_all_gap_in_msa = (cnt_not_aa < num_taxon)
        if ref_res not in not_aa:
            ref_i += 1
            if not_all_gap_in_msa:
                idmap[msa_i] = '{:03d}'.format(ref_i)
            ins_i = -1
        else:
            if not_all_gap_in_msa:
                ins_i += 1
                idmap[msa_i] = '{:03d}{}'.format(ref_i, codes[ins_i%len(codes)])
    return idmap

def idmap_format2raw(formated_seq, spacer='-'):
    return idmap_aln2raw(formated_seq, spacer)

def idmap_raw2format(formated_seq, spacer='-'):
    return idmap_raw2aln(formated_seq, spacer)

def idmap_aln2raw(formated_seq, spacer='-'):
    idmap = dict((v, k) for k, v in idmap_raw2aln(formated_seq).items())
    return idmap

def idmap_raw2aln(formated_seq, spacer='-'):
    accumulator = 0
    idmap = {}
    for i, aa in enumerate(formated_seq):
        if aa not in spacer:
            idmap[accumulator] = i
            accumulator += 1
    return idmap

def idmap_msa2msa(msa_a, msa_b, spacer='-'):
    """ msa_a and msa_b must be identical after removing empty/gappy columns """
    tmsa_a = transpose(msa_a)
    tmsa_b = transpose(msa_b)
    i_a, i_b = 0, 0
    idmap = {}
    while i_a < len(tmsa_a) and i_b < len(tmsa_b):
        if tmsa_a[i_a] == tmsa_b[i_b]: # identical columns
            idmap[i_a] = i_b
            i_a += 1
            i_b += 1
        else:
            # if the two columns are different, one must be gappy (thus removed in the another) 
            n_spacers_a = tmsa_a[i_a].count(spacer)
            n_spacers_b = tmsa_b[i_b].count(spacer)
            if n_spacers_a > n_spacers_b:
                i_a += 1
            else:
                i_b += 1
    return idmap

def transpose(msa):
    try:
        msa_t = []  # Transposed msa
        for j in range(len(msa[0])):
            msa_t.append(''.join(x[j] for x in msa))
    except KeyError:
        #print('Check whether the MSA is empty or not same lengh.', file=sys.stderr)
        assert(False)
    return msa_t

def cons(msa, unanimous=0.99, majority=0.5, use_most_common=False, ignore_space=False, breaking_aa_tie=False):
    msa_t = transpose(msa)
    aa_cnt = [collections.Counter(col) for col in msa_t]
    result = []
    for col in aa_cnt:
        cnt_all = sum(col.values())
        most_common_char, num = col.most_common(1)[0]
        if most_common_char == '-' and ignore_space and len(col) > 1:
            most_common_char, num = col.most_common(2)[1]

        # Break tie by frequency in proteins if there are other AA tie with the most_common AA
        if breaking_aa_tie:
            most_common_aas = [(aa, misc.AAFreq[aa]) if aa in misc.AAFreq else (aa, 0.0) for aa, cnt in col.items() if cnt == num] 
            most_common_char, aafreq = max(most_common_aas, key=lambda x: x[1])
        
        if 'X' in col:
            num += col['X']/20
        if num >= cnt_all * unanimous:
            result.append(most_common_char.upper())
        if num < cnt_all * unanimous and num >= cnt_all * majority:
            result.append(most_common_char.lower())
        if num < cnt_all * majority:
            if use_most_common:
                result.append(most_common_char.lower())
            else:
                result.append('?')
    return ''.join(result)

def mark_pngs(seq, merge_overlapped_sequon=False, spacer='-', marker='O'):
    pngs = set(get_pngs(seq, merge_overlapped_sequon, spacer))
    return ''.join(marker if i in pngs else aa for i, aa in enumerate(seq))

def get_pngs(seq, merge_overlapped_sequon=True, spacer='-', restrict_P3=False):
    """ resturn [i_pngs0, i_pngs1, ...]; i_pngs0, index of PNGS0 in seq """
    idmap = idmap_raw2format(seq, spacer)
    raw_seq = seq.upper()
    for c in spacer:
        raw_seq = raw_seq.replace(c, '')
    raw_seq += 'XXX'  # Append 'XXX' to handel sequence ends with 'N'
    pngs = set()
    overlapped = set()
    for i, aa in enumerate(raw_seq):
        if (aa.upper() == 'N'
            and raw_seq[i+1] != 'P'
            and raw_seq[i+2] in 'ST'):
            if restrict_P3 and raw_seq[i+3] == 'P':  #Ref: PMID 20510933
                continue
            pngs.add(i)
            if merge_overlapped_sequon and i-1 in pngs:
                overlapped.add(i-1) # NxT or second sequon is preferred: https://academic.oup.com/glycob/article/9/6/547/696392
    pngs -= overlapped
    pngs = [idmap[i] for i in pngs]
    pngs.sort()
    return pngs

def class_aa(seq):
    """ Adapted from: http://www.hiv.lanl.gov/content/sequence/ENTROPY/entropy_option.html
        #In input sequence       In Entropy calculations
        D and E                 a
        K and R                 b
        I and V                 i
        L and M                 l
        F and Y                 f (FYW -> f in HIV DB entropy)
        N (not PNGS) and Q      n
        N (PNGS)                g
        S and T                 s
    """
    # Potential N-linked glycosylation sites (PNGS):
    seq_classed = list(seq.upper())
    for pngs in get_pngs(seq):
        seq_classed[pngs] = 'g'
    seq_classed = ''.join(seq_classed)
    # Other sites
    trans_table = str.maketrans('DEKRIVLMFYNQST', 'aabbiillffnnss')
    return seq_classed.translate(trans_table)

def pssm_psic(msa, pseudocount=0.01, background=None, charset=misc.AAchar):
    """ return [{'A':score_a, 'C':score_c, ...}, {}, ...]"""
    msa_t = transpose(msa)
    col_count = []
    for i in range(len(msa_t)):
        col_count.append(psic(i, msa_t))

    if background is None:
        background = collections.defaultdict(lambda: 0.0)
        for col in col_count:
            for aa, cnt in col.items():
                background[aa] += cnt

    bg_all = sum(background[a] for a in charset)
    bg = dict((a, float(background[a])/bg_all) for a in charset)

    result = []
    for aa_cnt in col_count:
        cnt_all = sum(aa_cnt.values()) + pseudocount * len(charset)
        pos = {}
        for a in charset:
            p = float(aa_cnt[a] + pseudocount) / cnt_all
            pos[a] = math.log(p / bg[a])
        result.append(pos)
    return result

def pssm(msa, pseudocount=0.01, background=None, charset=misc.AAchar):
    """ return [{'A':score_a, 'C':score_c, ...}, {}, ...]"""
    if background is None:
        ## Use the input alignment as background
        #background = collections.Counter()
        #for seq in msa:
        #    background.update(seq)
        #print(background)
        # Use 1 as background
        background = dict((aa, 1) for aa in charset)
    bg_all = sum(background[a] for a in charset)
    bg = dict((a, float(background[a])/bg_all) for a in charset)
    #
    msa_t = transpose(msa)
    result = []
    for col in msa_t:
        aa_cnt = collections.Counter(col)
        cnt_all = sum(aa_cnt.values()) + pseudocount * len(charset)
        pos = {}
        for a in charset:
            p = float(aa_cnt[a] + pseudocount) / cnt_all
            pos[a] = math.log(p / bg[a])
        result.append(pos)
    return result

def pssm_w(msa, pseudocount=0.01, background=None, weights=None, charset=misc.AAchar):
    """ return [{'A':score_a, 'C':score_c, ...}, {}, ...]"""
    #
    if weights is None:
        weights = henikoff_weight(msa)
    #
    msa_t = transpose(msa)
    weighted_aa_cnt = weighted_cnt(msa_t, weights)
    if background is None:
        background = dict((aa, 0.0) for aa in charset)
        for aa_cnt in weighted_aa_cnt:
            for aa in charset:
                background[aa] += aa_cnt[aa]
    bg_all = sum(background[a] for a in charset)
    bg = dict((a, float(background[a])/bg_all) for a in charset)
    #
    result = []
    for aa_cnt in weighted_aa_cnt:
        cnt_all = sum(aa_cnt.values()) + pseudocount * len(charset)
        pos = {}
        for a in charset:
            p = float(aa_cnt[a] + pseudocount) / cnt_all
            pos[a] = math.log(p / bg[a])
        result.append(pos)
    return result

def score_pssm(pssm, sequence, weight=None, charset=misc.AAchar):
    score = 0
    if weight == None:
        weight = [1] * len(sequence)
    for pos, a, w in zip(pssm, sequence, weight):
        if a not in charset:  # penalty to Gap/X is same to the worst char
            score += min(pos.values()) * w
        else:
            score += pos[a] * w
    return score

def eliminate_gaps(msa, th=0.2, gap_chars='-', triplet=False):
    """Eliminate cols have high fraction of gaps

    Return: modified_msa, the alignment with gaps removed

    Keyword arguments:
    msa -- the input alignment, [s1, s2, s3, ...]
    th -- the threshold of gap ratio, above which the column will be removed
    gap_chars -- chars in the alignmen be considered as gaps
    """
    spacer = set(gap_chars)
    n_seqs = len(msa)
    msa_t = transpose(msa)
    step = 1 if not triplet else 3
    result_t = []
    for i in range(0, len(msa_t), step):
        n_gaps = 0.0
        for j in range(n_seqs):
            is_gap = all(msa_t[i+k][j] in spacer for k in range(step))
            n_gaps += 1.0 if is_gap else 0.0
        gap_ratio = n_gaps / n_seqs
        if gap_ratio < th:
            result_t += [msa_t[i+k] for k in range(step)]
    return transpose(result_t)

def na2aa(triplet, spacer='-'):
    try:
        return misc.NA2AA[triplet.upper()]
    except KeyError:
        if triplet.count(spacer) == 3:
            return '-'
        elif triplet.count(spacer) > 0:
            return '#'
        else:
            return 'X'

def aa3to1(resname3):
    try:
        return misc.AA3to1[resname3]
    except KeyError:
        return 'X'

def translate(na_seq):
    assert(len(na_seq) % 3 == 0)
    return ''.join(na2aa(triplet) for triplet in chunkstring(na_seq, 3))

def pepidx(aaseq, pep_len=9, spacer='-', not_aa='*#'):  # The index is corresponding to input seq
    '''return {pep:[(i, j), ...], ...}'''
    index = {}
    for i in range(len(aaseq) - pep_len + 1):
        if aaseq[i] not in spacer:
            acc = 1
            j = 1
            while acc < pep_len and i+j < len(aaseq):
                if aaseq[i+j] not in spacer:
                    acc += 1
                j += 1
            if acc == pep_len:
                pep = re.sub('[{}]'.format(spacer), '', aaseq[i:i+j])
                # Remove any pep that has not_aa in it
                if any((char in pep) for char in not_aa): continue
                if pep in index:
                    index[pep].append((i, i+j))
                else:
                    index[pep] = [(i, i+j)]
    return index

def group_consecutive_int(int_list):
    if int_list:
        int_list.sort()
        grouped = [[int_list[0]]]
        for i, j in zip(int_list[:-1], int_list[1:]):
            if j == i+1:
                grouped[-1].append(j)
            elif j > i+1:
                grouped.append([j])
            else:
                assert(False)
        return grouped
    else:
        return []

def weighted_cnt(msa_t, weight):
    """ Count AA of each column, with weight specified for each sequence """
    # The position_specific_independent_count() could replace weighted_cnt() if PSIC weight is prefered
    aa_cnt = []
    for col in msa_t:
        cnt = collections.defaultdict(lambda: 0)
        for i, aa in enumerate(col):
            cnt[aa] += weight[i]
        aa_cnt.append(cnt)
    return aa_cnt

def col_diversity(msa, eval_fun, weight=None, by_class=False, gap_char='-'):
    if by_class:
        msa_classed = [class_aa(seq) for seq in msa]
        msa_t = transpose(msa_classed)
    else: # Not by class
        msa_t = transpose(msa)
    if weight is None:
        weight = [1.0 for i in range(len(msa))]
    else:
        assert(len(weight) == len(msa))
    aa_cnt = weighted_cnt(msa_t, weight)
    result = []
    for col in aa_cnt:
        col_count = expand_ambiguos_aa_and_gap(col, by_class)
        prob = cnt2fractions(col_count)
        diversity_index = eval_fun(prob)
        result.append(diversity_index)
    return result

def shannon_entropy_index(fractions):
    se = sum( - p * math.log2(p) for p in fractions if p > 0)
    #se = sum( - p * math.log(p) for p in fractions if p > 0)
    return se

def n_golosov(fractions):
    p1 = max(fractions)
    n = sum(1.0/(1.0+(p1*p1/pi)-pi) for pi in fractions)
    return n

def shannon_entropy(msa, weight=None, by_class=False, gap_char='-'):
    return col_diversity(msa, shannon_entropy_index, weight, by_class, gap_char)

def effective_num_parties(msa, weight=None, by_class=False, gap_char='-'):
    return col_diversity(msa, n_golosov, weight, by_class, gap_char)

def henikoff_weight(msa, ignore_chars='-*#X'):
    """ Henikoff, S. & Henikoff, J.G. (1994) Position-based sequence weights. J. Mol. Biol. 243:574-578. """
    should_be_ignored = set(ignore_chars)
    msa_t = transpose(msa)
    weight = [0.0] * len(msa_t[0])
    for col in msa_t:
        aa_cnt = collections.Counter(col)
        if len(set(aa_cnt.keys()) & should_be_ignored) > 0:
            continue # ignore cols that is gappy or has abnormal chars
        n = len(aa_cnt)
        for k, v in aa_cnt.items():
            aa_cnt[k] = 1.0 / n / v
        for i, aa in enumerate(col):
            weight[i] += aa_cnt[aa]
    sum_w = sum(weight)
    weight = [w/sum_w for w in weight]
    return weight

def tree_weight(tree):
    """ Altschul, S.F., et al. (1989) Weights for data related by a tree. J. Mol. Biol. 207:647-653. """
    # tree is imported by Bio.Phylo
    def assign_r(clade):
        if clade.is_terminal():
            clade.r = clade.branch_length
        else:
            for c in clade.clades:
                if not hasattr(c, 'r'):
                    assign_r(c)
            if clade.branch_length is None:
                clade.branch_length = 0.0
            clade.r = clade.branch_length + (1.0/sum(1.0/c.r for c in clade.clades))
    def distribute_w(clade, w):
        if not hasattr(clade, 'w'):
            clade.w = w
        for c in clade.clades:
            c.w = w * (clade.r - clade.branch_length) / c.r
            distribute_w(c, c.w)
    clade = copy.deepcopy(tree.clade)
    assign_r(clade)
    distribute_w(clade, clade.count_terminals())
    w = {}
    for leave in clade.get_terminals():
        w[leave.name] = leave.w
    return w

def shannon_entropy_of_sites(msa, sites, w_seq=None, w_pos=None, by_class=False):
    """ Shannon Entropy of a set of sites

    sum(Hi) - sum_i,j(I_i,j) [i<j<n]
    """
    #Initialization
    working_msa = [class_aa(s) for s in msa] if by_class else msa
    w_pos = [1.0 for x in sites] if w_pos is None else w_pos
    w_seq = [1.0 for s in working_msa] if w_seq is None else w_seq
    assert(len(w_pos) == len(sites))
    assert(len(w_seq) == len(working_msa))
    se = lambda cnt: shannon_entropy_index(cnt2fractions(cnt))
    # Get entropy
    entropy = {}
    for i in sites:
        col_count = weighted_cnt_i(working_msa, i, w_seq)
        col_count = expand_ambiguos_aa_and_gap(col_count, by_class)
        entropy[i] = se(col_count)
    # Get mutual entropy (joint entropy as bridge)
    mutual_entropy = {}
    pairs = [p for p in it.combinations(sites, 2)]
    s2pos = dict((s, i) for i, s in enumerate(sites))
    wp = {}
    for p in pairs:
        comb_count = weighted_cnt_p(working_msa, p, w_seq)
        comb_count = expand_ambiguos_aa_and_gap_for_combs(comb_count, by_class)
        joint_entropy = se(comb_count)
        #print(p)
        mutual_entropy[p] = sum(entropy[x] for x in p) - joint_entropy
        #print(sum(entropy[x] for x in p))
        if sum(entropy[x] for x in p) > 0:
            wp[p] = sum(w_pos[s2pos[x]]*entropy[x] for x in p)/sum(entropy[x] for x in p)
        else:
            wp[p] = 1.0
    # Shannon Entropy of the Epitope
    return sum(w_pos[s2pos[i]] * entropy[i] for i in sites) - sum(mutual_entropy[p] * wp[p] for p in pairs)

def weighted_cnt_i(msa, i, w_seq):
    cnt = collections.defaultdict(lambda: 0.0)
    for s, w in zip(msa, w_seq):
        cnt[s[i]] += w
    return cnt

def weighted_cnt_p(msa, p, w_seq):
    cnt = collections.defaultdict(lambda: 0.0)
    for s, w in zip(msa, w_seq):
        cnt[tuple(s[i] for i in p)] += w
    return cnt

def cnt2fractions(cnt):
    acc_cnt = sum(cnt.values())
    assert(acc_cnt > 0)
    return [a/acc_cnt for a in cnt.values()]

def ignore_special_chars(cnt_col):
    special_chars = 'BZX'
    result = dict((k, v) for k, v in cnt_col.items() if k.isalpha() and k not in special_chars)
    return result

def ignore_special_chars_for_combs(cnt_comb):
    special_chars = 'BZX'
    result = {}
    for k, v in cnt_comb.items():
        if not (any(char in special_chars for char in k) or any(not char.isalpha() for char in k)):
            result[k] = v
    return result

def expand_ambiguos_aa_and_gap(cnt_col, by_class=False):
    """ Handeling ambiguos_aa in count list """
    # Ambiguos codes
    # IUPAC ambiguity codes for NA and AA (http://www.bioinformatics.org/sms2/iupac.html)
    if by_class:
        #'DEKRIVLMFYNQST' => 'aabbiillffnnss'
        ambiguos_aa = 'B:an Z:an X:GAsCilPfWHban'.split()
        uniq_chars = 'GAsCilPfWHban'
    else:
        ambiguos_aa = 'B:DN Z:EQ X:ACDEFGHIKLMNPQRSTVWY'.split()
        uniq_chars = 'GASTCVILMPFYWHKRDENQ'
    special_chars = dict(x.split(':') for x in ambiguos_aa)
    # Gaps: considered as X here
    gap_chars = '-.'
    special_chars.update(dict((g, uniq_chars) for g in gap_chars))
    #
    result = collections.defaultdict(lambda: 0.0)
    for aa, cnt in cnt_col.items():
        if aa in special_chars:
            potential_aas = special_chars[aa]
        else:
            potential_aas = aa
        for char in potential_aas:
            result[char] += float(cnt)/len(potential_aas)
    return result

def expand_ambiguos_aa_and_gap_for_combs(cnt_comb, by_class=False):
    """ Handeling ambiguos_aa in count list """
    # Ambiguos codes
    # IUPAC ambiguity codes for NA and AA (http://www.bioinformatics.org/sms2/iupac.html)
    if by_class:
        #'DEKRIVLMFYNQST' => 'aabbiillffnnss'
        ambiguos_aa = 'B:an Z:an X:GAsCilPfWHban'.split()
        uniq_chars = 'GAsCilPfWHban'
    else:
        ambiguos_aa = 'B:DN Z:EQ X:ACDEFGHIKLMNPQRSTVWY'.split()
        uniq_chars = 'GASTCVILMPFYWHKRDENQ'
    special_chars = dict(x.split(':') for x in ambiguos_aa)
    # Gaps: considered as X here
    gap_chars = '-.'
    special_chars.update(dict((g, uniq_chars) for g in gap_chars))
    #
    result = collections.defaultdict(lambda: 0.0)
    orignal_keys = tuple(cnt_comb.keys())
    for comb in orignal_keys:
        cnt = cnt_comb[comb]
        potential_aas = tuple(special_chars[aa] if aa in special_chars else aa  for aa in comb)
        potential_combs = tuple(expaneded_comb for expaneded_comb in it.product(*potential_aas))
        for expaneded_comb in potential_combs:
            result[expaneded_comb] += cnt/len(potential_combs)
    return result

def psic(site, msa_t):
    """ position-specific counts of independent observations

    Ref:
    Sunyaev, S.R., et al. (1999) PSIC: profile extraction from sequence alignments with
    position-specific counts of independent observations. Protein Eng. 12:387-394.
    """
    gap_chars = '-.'
    col = msa_t[site]
    aa_cnt = collections.defaultdict(lambda: 0.0)
    uniq_aa = set(col)
    for aa in uniq_aa:
        idx = [k for k, aa_k in enumerate(col) if aa_k == aa]
        independent_counts = []
        for j, col_j in enumerate(msa_t):
            if j != site:
                uniq_chars = set(col_j[k] for k in idx)
                if len(uniq_chars) == 1 and misc.first(uniq_chars) in gap_chars:
                    continue  # skip gap only columns
                independent_counts.append(len(uniq_chars))
        independent_count = sum(independent_counts) / len(independent_counts)
        aa_cnt[aa] = independent_count
    return aa_cnt

def psic_of_comb(comb, msa_t):
    gap_chars = '-.'
    idx_of_seqs = range(len(next(iter(msa_t))))
    get_aa_comb = lambda k, comb: tuple(msa_t[i][k] for i in comb)
    aa_combs = collections.Counter(get_aa_comb(k, comb) for k in idx_of_seqs)
    aa_comb_cnt = {}
    for aa_comb, cnt in aa_combs.items():
        idx = [k for k in idx_of_seqs if get_aa_comb(k, comb) == aa_comb]
        independent_counts = []
        for n, col_n in enumerate(msa_t):
            if n not in comb:
                uniq_chars = set(col_n[k] for k in idx)
                if len(uniq_chars) == 1 and next(iter(uniq_chars)) in gap_chars:
                    continue  # skip gap only columns
                independent_counts.append(len(uniq_chars))
        independent_count = sum(independent_counts) / len(independent_counts)
        aa_comb_cnt[aa_comb] = independent_count
    return aa_comb_cnt

# Tests
def test_psic():
    seq = ["AQAAD",
           "AQQQX",
           "QAAAD",
           "QAQQD"]
    msa_t = transpose(seq)
    print(psic(1, msa_t))
    print(psic_of_comb((0,1), msa_t))
    print(psic_of_comb((1,2), msa_t))
    se = lambda cnt: shannon_entropy_index(cnt2fractions(cnt))
    print(se(psic_of_comb((0,1), msa_t)))
    print(se(psic_of_comb((1,2), msa_t)))
    cnt0 = psic_of_comb((1,4), msa_t)
    print(cnt0)
    print(se(cnt0))
    cnt1 = expand_ambiguos_aa_and_gap_for_combs(cnt0)
    print(cnt1)
    print(se(cnt1))
    cnt2 = ignore_special_chars_for_combs(cnt0)
    print(cnt2)
    print(se(cnt2))
    print(msa_t)
    print(shannon_entropy_of_sites(seq, (0, 1, 2, 3, 4)))
    print(shannon_entropy(seq), sum(shannon_entropy(seq)))

def test_expand_ambiguos_aa_and_gap_for_combs():
    combs = collections.defaultdict(lambda: 0.0)
    combs[('A', 'B')] = 3
    combs[('Z', 'B')] = 3
    combs[('E', 'D')] = 1
    combs = expand_ambiguos_aa_and_gap_for_combs(combs)
    print(combs)
    combs = collections.defaultdict(lambda: 0.0)
    combs[('A', 'B')] = 3
    combs[('Z', 'B')] = 3
    combs[('a', 'a')] = 1
    combs = expand_ambiguos_aa_and_gap_for_combs(combs, True)
    print(combs)

if __name__ == '__main__':
    test_psic()
    test_expand_ambiguos_aa_and_gap_for_combs()
