#!/usr/bin/env python

import sys
import argparse
import collections
import math
from string import ascii_uppercase
from enum import Enum

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.text import TextPath
from matplotlib.path import Path
from matplotlib.offsetbox import AnnotationBbox, AuxTransformBox
from matplotlib.font_manager import FontProperties
from matplotlib.ticker import (MultipleLocator, AutoMinorLocator)

# Say, "the default sans-serif font is Arial "
mpl.rcParams['font.sans-serif'] = "Arial"
# Then, "ALWAYS use sans-serif fonts"
mpl.rcParams['font.family'] = "sans-serif"

class ColorOpt(Enum):
    Text = 1
    Background = 2
    Same = 3

ColorSpe = collections.namedtuple('ColorSpe', 'fg bg')

class CharBlock:
    def __init__(self,
                 color_scheme={}, fg='k', bg='w',
                 color_opt=ColorOpt.Text,
                 font_family='DejaVu Sans Mono',
                 **font_properties):
        """ Recommended font family: 'Bitstream Vera Sans Mono', 'DejaVu Sans Mono', or 'Courier New' """
        self.fp = FontProperties(family=font_family, **font_properties)
        self.default_bg = 'w'
        self.default_fg = 'k'
        self.cs = collections.defaultdict(lambda: ColorSpe(fg, bg))
        for k, v in color_scheme.items():
            if color_opt == ColorOpt.Same:
                self.cs[k]  = ColorSpe(v, v)
            elif color_opt == ColorOpt.Text:
                self.cs[k]  = ColorSpe(v, bg) 
            else:  # default: ColorOpt.Background
                self.cs[k]  = ColorSpe(fg, v)
        self.cache_path= {}

    def put(self, ax, char, x, y, height=1.0):
        if self.cs[char].bg != self.default_bg:
            ax.add_patch(patches.Rectangle((x-0.5, y), 1.0, height, ec='none', fc=self.cs[char].bg))
        if char.isalpha() and self.cs[char].fg != self.cs[char].bg:
            # Get the scales
            exp_bbox = patches.Rectangle((0, 0), 1.0, height).get_extents()
            if char not in self.cache_path:
                text_path = TextPath((0, 0), char, size=1.0, prop=self.fp)
                aa_patch = patches.PathPatch(text_path)
                aa_box = aa_patch.get_extents()
                self.cache_path[char] = (text_path, aa_box)
            else:
                text_path, aa_box = self.cache_path[char]
            y_scale = (exp_bbox.y1-exp_bbox.y0)/(aa_box.y1-aa_box.y0) * 0.98
            x_scale = (exp_bbox.x1-exp_bbox.x0)/(aa_box.x1-aa_box.x0) * 0.95
            #
            t = mpl.transforms.Affine2D().scale(x_scale, y_scale) + ax.transData
            aa_patch = patches.PathPatch(text_path, ec='none', fc=self.cs[char].fg, transform=t)
            offsetbox = AuxTransformBox(t)
            offsetbox.add_artist(aa_patch)
            ab = AnnotationBbox(offsetbox, (x, y+height/2.0), box_alignment=(0.5, 0.5), frameon=False)
            ax.add_artist(ab)

class ColorScheme(dict):
    def __init__(self, instream=None):
        if instream is None:
            color_scheme = '''
                #char RGB
                A #80bfff
                R #d91a00
                N #00bf00
                D #ff59ff
                C #ff9999
                Q #00a600
                E #cc00cc
                G #ffa64d
                H #00e0e0
                I #409fff
                L #2693ff
                K #ff4b33
                M #0d86ff
                F #0073e6
                P #d9d900
                S #00f200
                T #00d900
                W #0066cc
                Y #00c0c0
                V #59acff
                B #cccccc
                Z #cccccc
                X #cccccc
                O #00ffff
            '''
            instream = color_scheme.split('\n')
        for char, color in self.load_colors(instream).items():
            self[char] = color

    def load_colors(self, instream):
        color_scheme = {}
        for line in instream:  # misc.cleaned() consider hex number as comment, so it cannot be used here.
            cleaned = line.strip()
            if cleaned.startswith('#') or len(cleaned) == 0:
                continue
            else:
                char, rgb = cleaned.split()
                color_scheme[char] = rgb
        return color_scheme

    def hex_to_rgb(value):
        value = value.lstrip('#')
        lv = len(value)
        return tuple(int(value[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))

    def rgb_to_hex(rgb):
        return '#%02x%02x%02x' % rgb

def read_fasta(filestream):
    headers = []
    sequences = []
    for line in filestream.readlines():
        if type(line) is bytes:
            s = line.strip().decode()
        else:
            s = line.strip()
        if s.startswith('>'):
            headers.append(s[1:])
            sequences.append([])
        else:
            sequences[-1].append(s)
    sequences = [''.join(seqs).replace(' ', '') for seqs in sequences]
    return (headers, sequences)

def get_logostacks(msa, sites=None, n_charset=None):
    stacks = []
    # if sites not specified, use all sites in the msa
    if sites is None:
        logo_sites = [i for i in range(len(msa[0]))]
    else:
        logo_sites = sites
    # ref - PMID: 3525846; https://en.wikipedia.org/wiki/Sequence_logo
    if n_charset is None:
        n_charset = len(set(seq[i] for i in sites for seq in msa))
    n_seqs = len(msa)
    e_n = 1.0/math.log(2) * (n_charset - 1) / (2 * n_seqs)
    for i in logo_sites:
        cnt = collections.Counter(seq[i] for seq in msa)
        freq = dict((char, n_char/n_seqs) for char, n_char in cnt.items())
        H_i = -1.0 * sum(f * math.log2(f) for f in freq.values())
        R_i = math.log2(n_charset) - (H_i + e_n)
        stack = [(char, f * R_i) for char, f in freq.items()]
        stack.sort(key=lambda x: x[1])
        stacks.append(stack)
    return stacks

def get_locations(sites):
    locations = []
    groups = group_consecutive_int(sites)
    offset = 0
    for group in groups:
        locations += [i+offset for i in range(len(group))]
        offset += len(group) + 1
    return locations

def group_consecutive_int(int_list):
    if int_list:
        int_list.sort()
        grouped = [[int_list[0]]]
        for i, j in zip(int_list[:-1], int_list[1:]):
            if j == i+1:
                grouped[-1].append(j)
            elif j > i+1:
                grouped.append([j])
            else:
                assert(False)
        return grouped
    else:
        return []

def put_logostacks(ax, locations, logostacks, xlabels=None, text_formatter=CharBlock()):
    max_height = 0
    for x, logostack in zip(locations, logostacks):
        acc = 0
        for char, height in logostack:
            text_formatter.put(ax, char, x, acc, height)
            acc += height 
        if max_height < acc: max_height = acc
    #
    ax.set_ylim([0, max_height])
    ax.yaxis.set_minor_locator(MultipleLocator(0.2))
    ax.yaxis.set_major_locator(MultipleLocator(1.0))
    #
    x_ticks = locations
    ax.set_xlim([min(x_ticks)-0.5, max(x_ticks)+0.5])
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.set_xticks(x_ticks)
    if xlabels:
        ax.set_xticklabels(xlabels, rotation=90)
    else:
        ax.set_xticklabels([x+1 for x in x_ticks], rotation=90)

def autosize(n_sites):
    w_char = 0.3 
    x_pad = 1.0
    y_size = 3.0
    canvas_size = (x_pad + n_sites * w_char, y_size)
    return canvas_size

def parse():
    parser = argparse.ArgumentParser(description='Generate seqence logo plot')
    parser.add_argument('sequence', help='Input sequence alignment file (fasta format)', type=argparse.FileType('rt'))
    parser.add_argument('output', help='Output (seqlogo figure, *.png, *.pdf, etc)')
    parser.add_argument('-s', '--sites', help='Sites to be considered for the sequence logo plot (example: "1-5,10,12-17", default: all)')
    parser.add_argument('-cs', '--color_scheme', help='Color scheme for nucleotide or amino acid characters', type=argparse.FileType('rt'))
    parser.add_argument('--size', help='Figure size (in inch, example: 8.0x6.0, default: auto)')
    hl = parser.add_mutually_exclusive_group()
    hl.add_argument('--color_text', help='Color text (default)', action="store_const", dest="hl_style", const=ColorOpt.Text)
    hl.add_argument('--color_bg', help='Color background', action="store_const", dest="hl_style", const=ColorOpt.Background)
    parser.set_defaults(hl_style=ColorOpt.Text)
    return parser.parse_args()

def main():
    para = parse()
    #
    headers, msa = read_fasta(para.sequence)
    #
    if para.color_scheme is None:
        color_scheme = ColorScheme()
    else:
        color_scheme = ColorScheme(para.color_scheme)
    formatter = CharBlock(color_scheme, color_opt=para.hl_style)
    #
    if para.sites:
        sites = []
        for interval in para.sites.split(','):
            start_end = interval.split('-')
            s = int(start_end[0]) - 1
            e = int(start_end[-1])
            sites += [i for i in range(s, e)]
    else:
        sites = [i for i in range(len(msa[0]))]
    logostacks = get_logostacks(msa, sites)
    locations = get_locations(sites)
    xlabels = [i+1 for i in sites]
    #
    if para.size is None:  # automatic
        canvas_size = autosize(len(sites))
    else:
        w, h = para.size.split('x')
        canvas_size = (float(w), float(h)) 
    #
    fig, ax = plt.subplots(1, 1, figsize=canvas_size)
    put_logostacks(ax, locations, logostacks, xlabels, text_formatter=formatter)
    ax.set_ylabel('bits')
    plt.tight_layout()
    plt.savefig(para.output)
    #plt.show()

if __name__ == "__main__":
    main()
