import os.path
import collections
import re
import inspect
import math
import itertools as it

import scipy.cluster.hierarchy as hac
from scipy.spatial.distance import squareform
import numpy

import Bio.PDB
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))

try:
    from . import misc
    from ._wrap_dots_asa import dots_asa, dots_depth, dots_depth2
except Exception:
    import misc
    from _wrap_dots_asa import dots_asa, dots_depth, dots_depth2

def copy_model(model):
    print("copy_model() is not recommended (2019), use model.copy() instead(model: instance of Bio.PDB.Model)")
    return model.copy()

def copy_chains(struture, chains):
    extracted = struture.copy()
    for model in Bio.PDB.Selection.unfold_entities(extracted, 'M'):
        chains_to_be_removed = [chain.id for chain in model if chain.id not in chains]
        for chain_id in chains_to_be_removed:
            model.detach_child(chain_id)
    return extracted 

def parse_xyzrn(infile):
    xyz = []; r = []; n = []
    for line in infile.readlines():
        try:
            elements = line.strip().split()
            xyz.append([float(x) for x in elements[:3]])
            r.append(float(elements[3]))
            n.append(elements[-1])
        except ValueError:
            print("It should not happen")
    return (xyz, r, n)

def get_sc_centers(residues):
    """ Get center of side chain (CA is considered as side chain atom) """
    xyzs = []
    mc_atoms = set('C N O'.split())
    for r in residues:
        sc_coords = [atom.get_coord() for atom in r if atom.name not in mc_atoms]
        sc_enter = numpy.mean(sc_coords, axis=0)
        xyzs.append(sc_enter)
    return xyzs

def get_cb_xyzs(residues):
    xyzs = []
    for r in residues:
        if 'CB' in r:  # Use CB atoms when possible.
            xyzs.append(r['CB'].get_coord())
        elif 'CA' in r:  # If it is not possible (e.g. GLY, missing CB in resolved structure, etc), use CA atoms
            xyzs.append(r['CA'].get_coord())
        else:  # If both 'CA' and 'CB' are missing, use the center 
            coor = [0.0, 0.0, 0.0]
            n = len(r)
            for atom in r:
                current_coor = atom.get_coord()
                coor = [x+x_i/n for x, x_i in zip(coor, current_coor)]
            xyzs.append(coor)
            #xyzs.append(np.mean([atom.get_coord() for atom in r], axis=0))
    return xyzs

class XYZRN:
    def __init__(self):
        """ Parse atmtypenumber """
        self._r_united_atom = {}
        self._r_explicit = {}
        self._pat_and_type = []
        with open(currentdir + "/" + "data/atmtypenumber.txt") as infile:
            for line in misc.cleaned(infile):
                if line.startswith('radius'):
                    fields = line.split()
                    tag, atmtype, bond_dist, r_explicit = fields[:4]
                    self._r_explicit[atmtype] = float(r_explicit)
                    if len(fields) == 5:
                        self._r_united_atom[atmtype] = float(fields[4])
                else:
                    rname, aname, atmtype = line.split()
                    rpat = re.compile('^\s*{}\s*$'.format(rname.replace('*', '.*')))
                    apat = re.compile('^\s*{}\s*$'.format(aname.replace('\\', '')))
                    self._pat_and_type.append((rpat, apat, atmtype))

    def deduce(self, model, include_het=True):
        """ Get XYZRN of atoms in model """
        xyz, r, names = [], [], []
        for atom in model.get_atoms():  # if there are alternative locations, 
            if atom.is_disordered():
                try:
                    atom.disordered_select('A')
                except:
                    #print(atom.get_full_id())
                    pass
            atmid = atom.get_full_id() # atmid: (struname, model_number, chainid, (hetflag, resid, icode), (aname, altloc))
            if atmid[4][1] not in ' A': continue
            if not include_het and atmid[3][0] != ' ': continue
            xyz.append(atom.coord)
            rname = atom.parent.resname
            aname = atom.name
            r.append(self._r_by_atom_and_res(aname, rname))
            outid = encode_id(atmid)
            names.append(outid)
        return (xyz, r, names)

    def _r_by_atom_and_res(self, atmname, rname):
        for (rpat, apat, atmtype) in self._pat_and_type:
            if (apat.match(atmname) is not None) and (rpat.match(rname) is not None):
                if atmtype in self._r_united_atom:
                    r = self._r_united_atom[atmtype]
                elif atmtype in self._r_explicit:
                    r = self._r_explicit[atmtype]
                else:
                    r = 1.5
                break
        return r

def get_xyzrn(model, include_het=True):
    xyzrn = XYZRN()
    return xyzrn.deduce(model, include_het)

def get_depth(model, r_probe=15.0, d_ext=25.0, remove_interior_cavities=False, smpl_dots=None, include_het=True, plane_o=[], plane_n=[]):
    if smpl_dots is None:
        smpl_dots = misc.smpl256
    xyz, r, ids = get_xyzrn(model, include_het=include_het)
    depth = dots_depth(xyz, r, smpl_dots, r_probe, d_ext, plane_o, plane_n, remove_interior_cavities)
    depths = dict((x_id, a) for x_id, a in zip(ids, depth)) # depths: {AtomId:depth, ...}
    return depths

def get_asa(model, smpl_dots=None, probe=1.5, include_het=True, remove_interior_cavities=False):
    """ Get ASA area """
    if smpl_dots is None:
        smpl_dots = misc.smpl256
    xyz, r, ids = get_xyzrn(model, include_het=include_het)
    area = dots_asa(xyz, r, smpl_dots, probe, remove_interior_cavities)  # area: ASA of atoms, corresponding to order in xyz and r
    areas = dict((atom_id, a) for atom_id, a in zip(ids, area)) # areas: {AtomId:ASA, ...}
    return areas

def atom2res(struture, x_by_atom, f=sum):
    valid_res = misc.AAs|misc.Glycans
    residues = [r for r in Bio.PDB.Selection.unfold_entities(struture, 'R') if r.resname in valid_res]
    x_by_res = collections.OrderedDict()
    for res in residues:
        temp_rec = []
        for atom in res:
            atom_id = encode_id(atom.get_full_id())
            if atom_id in x_by_atom:
                temp_rec.append(x_by_atom[atom_id])
        if temp_rec:
            x_by_res[encode_id(res.get_full_id())] = f(temp_rec)
    return x_by_res

def get_scasa(structure, areas, ca_as_bb=False):
    if ca_as_bb:
        bbatoms = 'C N O CA'.split()
    else:
        bbatoms = 'C N O'.split()
    atom_name = lambda atom_id_str: atom_id_str.split(';')[-1].split(',')[0]
    sc_areas = dict((atom_id, asa) if atom_name(atom_id) not in bbatoms else (atom_id, 0) for atom_id, asa in areas.items())
    return atom2res(structure, sc_areas)

def area_by_residue(structure, areas, ca_as_bb=False):
    """ return [(res_id<Bio.PDB get_full_id()>, ASA, scASA), ...]"""
    asa_by_res = atom2res(structure, areas)
    scasa = get_scasa(structure, areas, ca_as_bb)
    result = collections.OrderedDict()
    for res in Bio.PDB.Selection.unfold_entities(structure, 'R'):
        res_id = res.get_full_id()
        id_str = encode_id(res_id)
        result[res_id] = (asa_by_res[id_str], scasa[id_str])
    return result

def get_sqdist_mat(atoms_a, atoms_b):
    coordsa = numpy.array([atom.coord for atom in atoms_a])
    coordsb = numpy.array([atom.coord for atom in atoms_b])
    return sqdist(coordsa, coordsb)

def sqdist(xyza, xyzb):
    ''' Get the distance matrix between coords array xyza and xyzb.

    Input: 
        xyza: [[xa1, ya1, za1], [xa2, ya2, za2], ...]
        xyzb: [[xb1, yb1, zb1], [xb2, yb2, zb2], ...]

    Output:
        distmatrix: (an x bn)
        [[D_a1_b1, D_a1_b2, D_a1_b3, ..., D_a1_bn], 
         [D_a2_b1, D_a2_b2, D_a2_b3, ..., D_a2_bn], 
         .
         .
         .
         [D_an_b1, D_an_b2, D_an_b3, ..., D_an_bn], 
    '''
    sizea = xyza.shape[0]
    sizeb = xyzb.shape[0]
    mat_a = xyza.reshape(sizea, 1, 3)
    mat_a = mat_a.repeat(sizeb, axis=1)
    # mat_a:
    # [[[xa1, ya1, za1], [[xa1, ya1, za1], ...],
    #  [[xa2, ya2, za2], [[xa2, ya2, za2], ...], 
    #  .
    #  .
    #  .
    #  [[xan, yan, zan], [[xan, yan, zan], ...]]
    mat_b = xyzb.reshape(1, sizeb, 3)
    mat_b = mat_b.repeat(sizea, axis=0)
    # mat_b:
    # [[[xb1, yb1, zb1], [xb2, yb2, zb2], ...],
    #  [[xb1, yb1, zb1], [xb2, yb2, zb2], ...],
    #  .
    #  .
    #  .
    #  [[xb1, yb1, zb1], [xb2, yb2, zb2], ...]]
    dist = mat_a - mat_b
    dist = numpy.sum(dist * dist, axis=2)
    return dist

def select_by_dist(atoms, atoms_ref, threshold=4.0):
    """ select atom in "atoms" within "threshold" of "atoms_ref" """
    sqdist_mat = get_sqdist_mat(atoms, atoms_ref)
    sqdist_min = numpy.min(sqdist_mat, axis=1)
    threshold2 = threshold * threshold
    result = [(atom, min_dist) for atom, min_dist in zip(atoms, sqdist_min) if min_dist < threshold2]
    result.sort(key=lambda atom_dist_pair:atom_dist_pair[1])
    return [atom_dist_pair[0] for atom_dist_pair in result]

def get_by_full_id(entity, full_id):
    """ Get elment in entity by full_id (entity could S/M/C/R/A) """
    def _get_by_corresponding_level_id(entity, entry_id):
        if len(entry_id) > 1:
            return _get_by_corresponding_level_id(entity[entry_id[0]], entry_id[1:])
        else:
            if entity.level == 'R':
                return entity[entry_id[0][0]]
            else:
                return entity[entry_id[0]]

    corresponding_level_id = full_id[misc.EntLevels.index(entity.level)+1:]
    return _get_by_corresponding_level_id(entity, corresponding_level_id)

def rsa(one_char_aa, asa):
    return float(asa/misc.TotalASA[one_char_aa])

def filter_res(struture, whitelist=misc.Glycans|misc.AAs):
    for chain in Bio.PDB.Selection.unfold_entities(struture, 'C'):
        residues_to_be_removed = [res.id for res in chain if res.resname not in whitelist]
        for res_id in residues_to_be_removed:
            chain.detach_child(res_id)

def remove_h(struture):
    for residue in Bio.PDB.Selection.unfold_entities(struture, 'R'):
        atoms_to_be_removed = [atom.id for atom in residue if atom.element == 'H']
        for atom_id in atoms_to_be_removed:
            residue.detach_child(atom_id)  

def clean_stru(struture, whitelist=misc.Glycans|misc.AAs):
    filter_res(struture, whitelist)
    remove_h(struture)

def group_glycan(prot, sugar_set=misc.Glycans, bond_length_th=misc.BondThreshold):
    """ Group glycans in protein and return {NGS: glycan_list, ...}.

    Input:
    prot -- Bio.PDB stru/model/chain instance
    sugar_set -- list of sugar names
    bond_length_th -- lenght of bond to link two glycans together

    Output: {NGS_full_id: glycan_full_id_list, ...}, ordered Dictionary
    NGS_full_id -- N-linked glycosylation site (residue N), full_id by Bio.PDB.X.get_full_id()
    glycan_full_id_list -- A list of glycans attaced to NGS, full_id by Bio.PDB.X.get_full_id()
    """
    def find_ngs(atoms_prot, atoms_sugar, bond_length_th):
        """ Tested: get_nb_atoms() is slower than select_by_dist() """
        nd2_candidates = select_by_dist(atoms_prot, atoms_sugar, threshold=bond_length_th)
        ngs = None
        for atom in nd2_candidates:
            if atom.name == 'ND2':
                ngs = atom.parent
                break
        if ngs is None:
            glycans = set(a.parent.get_full_id() for a in atoms_sugar)
            glycan_names = ','.join(gid[-2] + ''.join(str(x).replace(' ', '') for x in gid[-1])
                                        for gid in glycans)
            #warnings.warn('No bonded ASN found for glycan(s): {}'.format(glycan_names))
        return ngs
    sugar_res = [res for res in prot.get_residues() if res.resname in sugar_set]
    groups = []
    if len(sugar_res) < 1:
        pass
    elif len(sugar_res) == 1:
        atoms_prot = [atom for atom in prot.get_atoms() if atom.parent.resname not in sugar_set]
        atoms_sugar = [atom for residue in sugar_res for atom in residue]
        ngs = find_ngs(atoms_prot, atoms_sugar, bond_length_th)
        if ngs is not None:
            groups.append((ngs, sugar_res))
    else:
        # Build distance matrix between glycan residues
        min_sqdist = lambda r_a, r_b: numpy.min(get_sqdist_mat(r_a.child_list, r_b.child_list))
        sqdist_mat = numpy.zeros((len(sugar_res), len(sugar_res)))
        for i in range(len(sugar_res)):
            for j in range(i+1, len(sugar_res)):
                sqdist_mat[i, j] = sqdist_mat[j, i] = min_sqdist(sugar_res[i], sugar_res[j])
        # Cluster glycans according to distance matrix
        pdist = squareform(sqdist_mat)  # transform distance matrix to condensed form
        z = hac.linkage(pdist, 'single')  # single linkage method to cluster all things together
        # extract cluster by a distance cutoff 2.0*2.0
        cls_ids = hac.fcluster(z, bond_length_th*bond_length_th, 'distance') 
        glycan_clusters = collections.defaultdict(list)
        for cls_id, sugar in zip(cls_ids, sugar_res):
            glycan_clusters[cls_id].append(sugar)
        # Get the NGS of each cluster
        atoms_prot = [atom for atom in prot.get_atoms() if atom.parent.resname not in sugar_set]
        for cls_id, glycans in glycan_clusters.items():
            atoms_sugar = [atom for residue in glycans  for atom in residue]
            ngs = find_ngs(atoms_prot, atoms_sugar, bond_length_th)
            if ngs is not None:
                groups.append((ngs, glycans))
    groups.sort(key=lambda x: x[0].id)
    result = collections.OrderedDict((ngs.get_full_id(), [glycan.get_full_id() for glycan in glycans]) for ngs, glycans in groups)
    return result  

def fix_glycan_ids(structure):
    """ The input structure will be modified, no return values """
    # Detect glycan groups
    glycan_groups = group_glycan(structure)
    glycan_groups_by_chains = collections.OrderedDict()
    for ngs, glycans in glycan_groups.items():
        asn_chain_id = ngs[2]
        if asn_chain_id not in glycan_groups_by_chains:
            glycan_groups_by_chains[asn_chain_id] = [(ngs, glycans)]
        else:
            glycan_groups_by_chains[asn_chain_id].append((ngs, glycans))
    # Check the glycan ids and modify chain ids, residue ids and segment ids
    for chain_id, glycan_groups in glycan_groups_by_chains.items():
        occupied_ids = set()
        for ngs, glycans in glycan_groups:
            asn_chainid = ngs[2]
            asn = get_by_full_id(structure, ngs)
            asn_segid = asn.get_segid()
            base = math.ceil(len(asn.parent.child_list)/1000)*1000
            offset = 0
            for j, glycan in enumerate(glycans):
                glycan_res = get_by_full_id(structure, glycan)
                glycan_segid = glycan_res.get_segid()
                glycan_chainid = glycan[2]
                if glycan_chainid != asn_chainid or asn_segid != glycan_segid:
                    revised_id = list(glycan_res.id)
                    new_sn = base + ngs[3][1] + j + offset
                    while new_sn in occupied_ids:
                        offset += 10
                        new_sn += 10
                    revised_id[1] = new_sn
                    occupied_ids.add(new_sn)
                    # Carry out the modifications
                    structure[glycan_chainid].detach_child(glycan_res.id)
                    glycan_res.id = tuple(revised_id)
                    glycan_res.segid = asn_segid
                    structure[asn_chainid].add(glycan_res)
    # There is no return here, structure is modified inplace

def encode_id(x_id):
    encoded = ';'.join(','.join(str(y) for y in x) if type(x) == tuple else str(x) for x in x_id)
    return f'({encoded})'

def encode_resid(resid):
    return encode_id(resid)

def decode_id(id_str):
    decoded = id_str.strip('()').split(';')
    decoded[1] = int(decoded[1])
    if len(decoded) >=4:
        decoded[3] = decoded[3].split(',')
        decoded[3][1] = int(decoded[3][1])
        decoded[3] = tuple(decoded[3])
    if len(decoded) == 5:
        decoded[4] = tuple(decoded[4].split(','))
    return tuple(decoded)

def decode_resid(id_str):
    return decode_id(id_str)

def get_nb_atoms(atoms, atoms_ref, threshold=4.0):
    """ Retrive IDs of atoms in "atoms_ref" within "threshold" of any atom in "atoms"

    Input:
    atoms -- atom whoes neighbor should be retrived, list of Bio.PDB atom instances
    atoms_ref -- neighbor atoms are subject to extraction, list of Bio.PDB atom instances
    threshold -- square of distance threshold to define neighbor atoms

    Output: {atom_id: id_nb_atom_in_ref, ...}
    atom_id  -- id of atoms have neighbors in atoms_ref
    id_nb_atom_in_ref -- list of ids of neighbor atoms in atoms_ref
    """
    id2coord = dict((atom.get_full_id(), atom.coord) for atom in atoms_ref)
    ref_frame = CubicIdx(atoms_ref, threshold)
    result = collections.OrderedDict()
    threshold2 = threshold * threshold
    result = {}
    for atom in atoms:
        potential_nb_ids = ref_frame.query(atom.coord, threshold)
        if len(potential_nb_ids) > 0:
            potential_nb_coords = numpy.array([id2coord[atom_id] for atom_id in potential_nb_ids])
            d_diff = potential_nb_coords - atom.coord
            dist2 = numpy.sum(d_diff * d_diff, axis=1)
            nb_ids = [atom_id for d2, atom_id in  zip(dist2, potential_nb_ids) if d2 < threshold2]
            if nb_ids:
                result[atom.get_full_id()] = nb_ids
    return result

class CubicIdx:
    """ Simple yet powerful spacial index """
    def __init__(self, atoms, grid):
        assert(grid > 0.0)
        self._grid = grid
        self._idx = collections.defaultdict(list)
        for atom in atoms:
            self._idx[self._index(atom.coord)].append(atom.get_full_id())

    def query(self, q_xyz, q_range):
        n = int(math.ceil(q_range/self._grid))
        search_range = (range(i-n, i+n+1) for i in self._index(q_xyz))
        search_grids = it.product(*search_range)
        result = sum((self._idx[grid] for grid in search_grids if grid in self._idx), [])
        return result

    def _index(self, xyz):
        return tuple(int(math.floor(x/self._grid)) for x in xyz)

def get_pep_chain_and_name(cif_dict):
    ''' Get chain id(s) and name of entitie in mmCIF dictionary'''
    entityids = cif_dict['_entity.id']
    if len(entityids) == 1:
        entityid_name_dict = {cif_dict['_entity.id']: cif_dict['_entity.pdbx_description']}
    else:
        entityid_name_dict = dict((x_id, name) for x_id, name in zip(cif_dict['_entity.id'], cif_dict['_entity.pdbx_description']))
    entityid_chain_dict = dict((x_id, chain) for x_id, chain in zip(cif_dict['_entity_poly.entity_id'], cif_dict['_entity_poly.pdbx_strand_id']))
    #import pdb; pdb.set_trace()
    return tuple((entityid_chain_dict[x_id], entityid_name_dict[x_id]) for x_id in entityids if x_id in entityid_chain_dict)

def convenient2rid(convenient_rid):
    """ [convenient_rid] '18', '32A' --> [rid] ('', 18, ''), ('', 32, 'A') """
    if convenient_rid[-1].isalpha():
        return (' ', int(convenient_rid[:-1]),  convenient_rid[-1])
    else:
        return (' ', int(convenient_rid),  ' ')

def extract_seg(structure, segnote):
    """ [segnote] ('C', '18', '32A') --> [res] [ <res18>, ..., <res32A> ] """
    chain_id, p1, p2 = segnote
    start = convenient2rid(p1)
    end = convenient2rid(p2)
    # Extract residues
    assert(start[1] <= end[1])
    residues = []
    for res in structure[0][chain_id]:
        if res.id == start or len(residues) > 0 or res.id == end:
            residues.append(res)
        if res.id == end:
            break
    return residues

#import tempfile
#import shutil
#import subprocess
#class DropHet(Bio.PDB.Select):
#    def accept_residue(self, residue):
#        if residue.id[0].startswith('H_'):
#            return 0
#        else:
#            return 1

#def run_msms(model, probe=1.5, include_het=True):
#    """ Get ASA area """
#    msms = "msms"
#    pdb2xyzr = currentdir + "/" + "../aux_bin/pdb2xyzrn"
#    #pdb2xyzr = "pdb2xyzrn"
#    pdbio = Bio.PDB.PDBIO()
#    output_stru = model.copy()
#    pdbio.set_structure(output_stru)  # set_structure() may modify model (add parent information)
#    tmpdir = tempfile.mkdtemp()  # temp dir for files
#    basename = 'model'
#    if include_het:
#        pdbio.save(tmpdir+'/'+ basename + '.pdb')
#    else:
#        pdbio.save(tmpdir+'/'+ basename + '.pdb', DropHet())
#    xyzrF = tmpdir + '/' + basename + '.xyzrn'
#    with open(xyzrF, 'wt') as xyzr:
#        subprocess.call([pdb2xyzr, tmpdir + '/' + basename + '.pdb'], stdout=xyzr)
#    areaF = tmpdir + '/' + basename + '.area'
#    with open ('/dev/null', 'w') as FNULL:
#        subprocess.call([msms, '-if', xyzrF, '-surface', 'ases', '-af', areaF, '-probe_radius', str(probe)],
#                        stdout=FNULL,
#                        stderr=FNULL)
#    areas = {}
#    with open(areaF, 'rt') as af:
#        # area file format: [(Atom#, SES, ASA, AtomID), ...]
#        for line in misc.cleaned(af):
#            atom_sn, ses, asa, atom_id = line.split()
#            areas[atom_id] = float(asa)
#    shutil.rmtree(tmpdir)
#    return areas

#def get_name_in_xyzrn(atom):
#    res = atom.parent
#    chain = res.parent
#    return '{}_{}_{}_{}_{}'.format(atom.name, res.resname, chain.id, res.id[1], res.id[2]).replace(' ', '')

if __name__ == '__main__':
    pass
