#!/usr/bin/env python

import sys
import argparse
import collections
import math
import re

import Bio.PDB
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from inspect import currentframe, getframeinfo
from pathlib import Path
filename = getframeinfo(currentframe()).filename
parent = Path(filename).resolve().parent.parent
sys.path.insert(0, str(parent))
from hb_lib import strulib, seqlib, misc

def parse():
    parser = argparse.ArgumentParser(description="Renumber residue ids according to reference sequence", formatter_class=misc.SmartFormatter)
    parser.add_argument('pdbfile', help='pdbfile (.cif/.pdb/.cif.gz/.pdb.gz)')
    parser.add_argument('msa', help='fasta format', type=argparse.FileType('rt'))
    parser.add_argument('-n', '--seq_name', help='reference sequence to be used to number residues in the alignment (default: first).', default=None)
    parser.add_argument('output', help='pdbfile (.pdb/.pdb.gz)')
    return parser.parse_args()

def main():
    para = parse()

    parser = Bio.PDB.PDBParser()  # default: PDB format
    if '.cif' in para.pdbfile:  # Not tested yet
        parser = Bio.PDB.MMCIFParser()
    with misc.zopen(para.pdbfile) as infile:
        structure = parser.get_structure('', infile)[0]

    msa_heads, msa_seqs = seqlib.read_fasta(para.msa)
    seq_dict = dict((n, s) for n, s in zip(msa_heads, msa_seqs))
    if para.seq_name:
        refseq = seq_dict[para.seq_name]
    else:
        refseq = msa_seqs[0]
    rawseq2id_msa = dict((seq.replace('-', ''), i) for i, seq in enumerate(msa_seqs))

    # 
    p_refid = re.compile(r"([0-9]+)([^0-9]*)")
    for chain in structure:
        stru_seq_raw = ''.join(seqlib.aa3to1(r.resname) for r in chain if r.resname in misc.AAs)
        stru_seq_aln = msa_seqs[rawseq2id_msa[stru_seq_raw]]
        iraw2msa = seqlib.idmap_raw2aln(stru_seq_aln)
        imsa2ref = seqlib.idmap_msa2ref([stru_seq_aln], refseq)
        # set to impossible ids
        for i, r in enumerate(chain):
            if i >= len(stru_seq_raw): break
            revised_id = list(r.id)
            revised_id[1] = i + 9999
            r.id = tuple(revised_id)
        # renumber all residues
        for i, r in enumerate(chain):
            if i >= len(stru_seq_raw): break
            revised_id = list(r.id)
            i_msa = iraw2msa[i]
            i_ref = imsa2ref[i_msa]
            res_sn, icode = p_refid.match(i_ref).groups()
            revised_id[1] = int(res_sn)
            revised_id[2] = icode if len(icode) > 0 else ' '
            r.id = tuple(revised_id)

    # Output
    with misc.zopen(para.output, 'wt') as outstream:
        io = Bio.PDB.PDBIO()
        io.set_structure(structure)
        io.save(outstream)


if __name__ == '__main__':
    main()
